#!/usr/bin/env python3

from binascii import unhexlify
from eth_account.datastructures import SignedTransaction
from web3.datastructures import AttributeDict
from web3.contract import Contract
from web3 import Web3
from web3.types import Wei
from typing import (
    Optional, Union
)
from base64 import b64encode

import json

from ...exceptions import (
    AddressError, NetworkError, UnitError
)
from ...utils import clean_transaction_raw
from ..config import xinfin as config
from .wallet import Wallet
from .htlc import HTLC
from .rpc import (
    get_web3, get_transaction_receipt
)
from .utils import (
    _AttributeDict, is_network, is_address, to_checksum_address, amount_unit_converter
)
from .solver import (
    FundSolver, WithdrawSolver, RefundSolver
)


class Transaction:
    """
    XinFin Transaction.

    :param network: XinFin network, defaults to ``mainnet``.
    :type network: str
    :param provider: XinFin network provider, defaults to ``http``.
    :type provider: str

    :returns: Transaction -- XinFin transaction instance.

    .. note::
        XinFin has only three networks, ``mainnet``, ``apothem`` and ``testnet``.
    """

    def __init__(self, network: str = config["network"], provider: str = config["provider"],
                 token: Optional[str] = None):

        # Check parameter instances
        if not is_network(network=network):
            raise NetworkError(f"Invalid XinFin '{network}' network",
                               "choose only 'mainnet', 'apothem' or 'testnet' networks.")

        self._network: str = network
        self.web3: Web3 = get_web3(
            network=network, provider=provider
        )

        self._transaction: Optional[dict] = None
        self._signature: Optional[dict] = None
        self._type: Optional[str] = None
        self._fee: Optional[Wei] = None

    def fee(self, unit: str = config["unit"]) -> Union[Wei, int, float]:
        """
        Get XinFin transaction fee.

        :param unit: XinFin unit, default to ``Wei``.
        :type unit: str

        :returns: Wei, int, float -- XinFin transaction fee.

        >>> from swap.providers.xinfin.htlc import HTLC
        >>> from swap.providers.xinfin.transaction import FundTransaction
        >>> from swap.utils import sha256, get_current_timestamp
        >>> htlc: HTLC = HTLC(contract_address="xdcdE06b10c67765c8C0b9F64E0eF423b45Eb86b8e7", network="testnet")
        >>> htlc.build_htlc(secret_hash=sha256("Hello Meheret!"), recipient_address="xdcf8D43806260CFc6cC79fB408BA1897054667F81C", sender_address="xdc2224caA2235DF8Da3D2016d2AB1137D2d548A232", endtime=get_current_timestamp(plus=3600))
        >>> fund_transaction: FundTransaction = FundTransaction(network="testnet")
        >>> fund_transaction.build_transaction(address="xdc2224caA2235DF8Da3D2016d2AB1137D2d548A232", htlc=htlc, amount=3, unit="XDC")
        >>> fund_transaction.fee(unit="Wei")
        138436
        """

        # Check transaction
        if not self._transaction:
            raise ValueError("Transaction is none, build transaction first.")

        if unit not in ["XDC", "Gwei", "Wei"]:
            raise UnitError(f"Invalid XinFin '{unit}' unit", "choose only 'XDC', 'Gwei' or 'Wei' units.")
        return self._fee if unit == "Wei" else \
            amount_unit_converter(amount=self._fee, unit_from=f"Wei2{unit}")

    def hash(self) -> Optional[str]:
        """
        Get XinFin transaction hash.

        :returns: str -- XinFin transaction hash.

        >>> from swap.providers.xinfin.transaction import WithdrawTransaction
        >>> from swap.providers.xinfin.solver import WithdrawSolver
        >>> withdraw_transaction: WithdrawTransaction = WithdrawTransaction(network="testnet")
        >>> withdraw_transaction.build_transaction(transaction_hash="0x0d4c93546aa3e5e476455931a63f1a97a2624e3b516e3fd8e3a582cb20aaeef9", secret_key="Hello Meheret!", address="xdcf8D43806260CFc6cC79fB408BA1897054667F81C", contract_address="xdcdE06b10c67765c8C0b9F64E0eF423b45Eb86b8e7")
        >>> withdraw_solver: WithdrawSolver = WithdrawSolver(xprivate_key="xprv9s21ZrQH143K4Kpce43z5guPyxLrFoc2i8aQAq835Zzp4Rt7i6nZaMCnVSDyHT6MnmJJGKHMrCUqaYpGojrug1ZN5qQDdShQffmkyv5xyUR", path="m/44'/550'/0'/0/0")
        >>> withdraw_transaction.sign(solver=withdraw_solver)
        >>> withdraw_transaction.hash()
        "0xe8e8738c791385738661573ad4de63dd81b77d240b6138ca476ea8cdcbb29a21"
        """

        # Check transaction
        if not self._transaction:
            raise ValueError("Transaction is none, build transaction first.")

        return self._signature["hash"] if self._signature else None

    def json(self) -> dict:
        """
        Get XinFin transaction fee.

        :returns: Wei, int, float -- XinFin transaction fee.

        >>> from swap.providers.xinfin.htlc import HTLC
        >>> from swap.providers.xinfin.transaction import FundTransaction
        >>> from swap.utils import sha256, get_current_timestamp
        >>> htlc: HTLC = HTLC(contract_address="xdcdE06b10c67765c8C0b9F64E0eF423b45Eb86b8e7", network="testnet")
        >>> htlc.build_htlc(secret_hash=sha256("Hello Meheret!"), recipient_address="xdcf8D43806260CFc6cC79fB408BA1897054667F81C", sender_address="xdc2224caA2235DF8Da3D2016d2AB1137D2d548A232", endtime=get_current_timestamp(plus=3600))
        >>> fund_transaction: FundTransaction = FundTransaction(network="testnet")
        >>> fund_transaction.build_transaction(address="xdc2224caA2235DF8Da3D2016d2AB1137D2d548A232", htlc=htlc, amount=3, unit="XDC")
        >>> fund_transaction.json()
        {'chainId': 1337, 'from': '0x2224caA2235DF8Da3D2016d2AB1137D2d548A232', 'value': 3000000000000000000, 'nonce': 2, 'gas': 138436, 'gasPrice': 20000000000, 'to': '0xdE06b10c67765c8C0b9F64E0eF423b45Eb86b8e7', 'data': '0xf4fd30623a26da82ead15a80533a02696656b14b5dbfd84eb14790f2e1be5e9e45820eeb000000000000000000000000f8d43806260cfc6cc79fb408ba1897054667f81c0000000000000000000000002224caa2235df8da3d2016d2ab1137d2d548a2320000000000000000000000000000000000000000000000000000000060e000d3'}
        """

        # Check transaction
        if not self._transaction:
            raise ValueError("Transaction is none, build transaction first.")

        return self._transaction

    def raw(self) -> Optional[str]:
        """
        Get XinFin transaction hash.

        :returns: str -- XinFin transaction hash.

        >>> from swap.providers.xinfin.transaction import RefundTransaction
        >>> from swap.providers.xinfin.solver import RefundSolver
        >>> refund_transaction: RefundTransaction = RefundTransaction(network="testnet")
        >>> refund_transaction.build_transaction(transaction_hash="0x0d4c93546aa3e5e476455931a63f1a97a2624e3b516e3fd8e3a582cb20aaeef9", address="xdc2224caA2235DF8Da3D2016d2AB1137D2d548A232", contract_address="xdcdE06b10c67765c8C0b9F64E0eF423b45Eb86b8e7")
        >>> refund_solver: RefundSolver = RefundSolver(xprivate_key="xprv9s21ZrQH143K3Y3pdbkbjreZQ9RVmqTLhRgf86uZyCJk2ou36YdUJt5frjwihGWmV1fQEDioiGZXWXUbHLy3kQf5xmhvhp8dZ2tfn6tgGUj", path="m/44'/550'/0'/0/0")
        >>> refund_transaction.sign(solver=refund_solver)
        >>> refund_transaction.hash()
        "0xf88a028504a817c80082e76094de06b10c67765c8c0b9f64e0ef423b45eb86b8e780a47249fbb61909575c436a0eabe6caa72d4feb2c4aeceef586fe94ca82f36ce9c20efda4b4820a95a05ed63e467fb541b728dc7253ea4f9c4f2ada130ef78ffaba8de9c5e92536ce42a034ba97172cb8726cdfbaba14b10a817a0c4c4bccc6d8f2a27fc1711752ed2ab2"
        """

        # Check transaction
        if not self._transaction:
            raise ValueError("Transaction is none, build transaction first.")

        return self._signature["rawTransaction"] if self._signature else None

    def type(self) -> str:
        """
        Get XinFin transaction hash.

        :returns: str -- XinFin transaction hash.

        >>> from swap.providers.xinfin.transaction import WithdrawTransaction
        >>> from swap.providers.xinfin.solver import WithdrawSolver
        >>> withdraw_transaction: WithdrawTransaction = WithdrawTransaction(network="testnet")
        >>> withdraw_transaction.build_transaction(transaction_hash="0x0d4c93546aa3e5e476455931a63f1a97a2624e3b516e3fd8e3a582cb20aaeef9", secret_key="Hello Meheret!", address="xdcf8D43806260CFc6cC79fB408BA1897054667F81C", contract_address="xdcdE06b10c67765c8C0b9F64E0eF423b45Eb86b8e7")
        >>> withdraw_transaction.type()
        "xinfin_withdraw_unsigned"
        """

        # Check transaction
        if not self._transaction:
            raise ValueError("Transaction is none, build transaction first.")

        return self._type

    def signature(self) -> dict:
        """
        Get XinFin transaction hash.

        :returns: str -- XinFin transaction hash.

        >>> from swap.providers.xinfin.transaction import RefundTransaction
        >>> from swap.providers.xinfin.solver import RefundSolver
        >>> refund_transaction: RefundTransaction = RefundTransaction(network="testnet")
        >>> refund_transaction.build_transaction(transaction_hash="0x0d4c93546aa3e5e476455931a63f1a97a2624e3b516e3fd8e3a582cb20aaeef9", address="xdc2224caA2235DF8Da3D2016d2AB1137D2d548A232", contract_address="xdcdE06b10c67765c8C0b9F64E0eF423b45Eb86b8e7")
        >>> refund_solver: RefundSolver = RefundSolver(xprivate_key="xprv9s21ZrQH143K3Y3pdbkbjreZQ9RVmqTLhRgf86uZyCJk2ou36YdUJt5frjwihGWmV1fQEDioiGZXWXUbHLy3kQf5xmhvhp8dZ2tfn6tgGUj", path="m/44'/550'/0'/0/0")
        >>> refund_transaction.sign(solver=refund_solver)
        >>> refund_transaction.signature()
        {'hash': '0x90449ab8e3736feae4980554bb129b408f88d0003e569022cf8e00817cc2a7d9', 'rawTransaction': '0xf88a028504a817c80082e76094de06b10c67765c8c0b9f64e0ef423b45eb86b8e780a47249fbb61909575c436a0eabe6caa72d4feb2c4aeceef586fe94ca82f36ce9c20efda4b4820a95a05ed63e467fb541b728dc7253ea4f9c4f2ada130ef78ffaba8de9c5e92536ce42a034ba97172cb8726cdfbaba14b10a817a0c4c4bccc6d8f2a27fc1711752ed2ab2', 'r': 42895942847608608192932856733711858695420995837709512084644654454168196927042, 's': 23849944468865388317715121201379699989753020274517556595896033163737398323890, 'v': 2709}
        """

        # Check transaction
        if not self._transaction:
            raise ValueError("Transaction is none, build transaction first.")

        return self._signature

    def transaction_raw(self) -> str:
        """
        Get XinFin fund transaction raw.

        :returns: str -- XinFin fund transaction raw.

        >>> from swap.providers.xinfin.htlc import HTLC
        >>> from swap.providers.xinfin.transaction import FundTransaction
        >>> from swap.utils import sha256, get_current_timestamp
        >>> htlc: HTLC = HTLC(contract_address="xdcdE06b10c67765c8C0b9F64E0eF423b45Eb86b8e7", network="testnet")
        >>> htlc.build_htlc(secret_hash=sha256("Hello Meheret!"), recipient_address="xdcf8D43806260CFc6cC79fB408BA1897054667F81C", sender_address="xdc2224caA2235DF8Da3D2016d2AB1137D2d548A232", endtime=get_current_timestamp(plus=3600))
        >>> fund_transaction: FundTransaction = FundTransaction(network="testnet")
        >>> fund_transaction.build_transaction(address="xdc2224caA2235DF8Da3D2016d2AB1137D2d548A232", htlc=htlc, amount=3, unit="XDC")
        >>> fund_transaction.transaction_raw()
        "eyJmZWUiOiAxMzg0MzYsICJ0eXBlIjogInhpbmZpbl9mdW5kX3Vuc2lnbmVkIiwgInRyYW5zYWN0aW9uIjogeyJjaGFpbklkIjogMTMzNywgImZyb20iOiAiMHgyMjI0Y2FBMjIzNURGOERhM0QyMDE2ZDJBQjExMzdEMmQ1NDhBMjMyIiwgInZhbHVlIjogMzAwMDAwMDAwMDAwMDAwMDAwMCwgIm5vbmNlIjogMiwgImdhcyI6IDEzODQzNiwgImdhc1ByaWNlIjogMjAwMDAwMDAwMDAsICJ0byI6ICIweGRFMDZiMTBjNjc3NjVjOEMwYjlGNjRFMGVGNDIzYjQ1RWI4NmI4ZTciLCAiZGF0YSI6ICIweGY0ZmQzMDYyM2EyNmRhODJlYWQxNWE4MDUzM2EwMjY5NjY1NmIxNGI1ZGJmZDg0ZWIxNDc5MGYyZTFiZTVlOWU0NTgyMGVlYjAwMDAwMDAwMDAwMDAwMDAwMDAwMDAwMGY4ZDQzODA2MjYwY2ZjNmNjNzlmYjQwOGJhMTg5NzA1NDY2N2Y4MWMwMDAwMDAwMDAwMDAwMDAwMDAwMDAwMDAyMjI0Y2FhMjIzNWRmOGRhM2QyMDE2ZDJhYjExMzdkMmQ1NDhhMjMyMDAwMDAwMDAwMDAwMDAwMDAwMDAwMDAwMDAwMDAwMDAwMDAwMDAwMDAwMDAwMDAwMDAwMDAwMDA2MGUwMDBkMyJ9LCAic2lnbmF0dXJlIjogbnVsbCwgIm5ldHdvcmsiOiAidGVzdG5ldCJ9"
        """

        # Check transaction
        if not self._transaction:
            raise ValueError("Transaction is none, build transaction first.")

        return clean_transaction_raw(b64encode(str(json.dumps(dict(
            fee=self._fee,
            type=self._type,
            transaction=self._transaction,
            signature=self._signature,
            network=self._network
        ))).encode()).decode())


class FundTransaction(Transaction):
    """
    XinFin Fund transaction.

    :param network: XinFin network, defaults to ``mainnet``.
    :type network: str
    :param provider: XinFin network provider, defaults to ``http``.
    :type provider: str

    :returns: FundTransaction -- XinFin fund transaction instance.

    .. warning::
        Do not forget to build transaction after initialize fund transaction.
    """

    def __init__(self, network: str = config["network"], provider: str = config["provider"]):
        super().__init__(
            network=network, provider=provider
        )

    def build_transaction(self, address: str, htlc: HTLC, amount: Union[Wei, int, float],
                          unit: str = config["unit"]) -> "FundTransaction":
        """
        Build XinFin fund transaction.

        :param htlc: XinFin HTLC instance.
        :type htlc: xinfin.htlc.HTLC
        :param address: XinFin sender address.
        :type address: str
        :param amount: XinFin amount.
        :type amount: Wei, int, float
        :param unit: XinFin unit, default to ``Wei``.
        :type unit: str

        :returns: FundTransaction -- XinFin fund transaction instance.

        >>> from swap.providers.xinfin.htlc import HTLC
        >>> from swap.providers.xinfin.transaction import FundTransaction
        >>> from swap.utils import sha256, get_current_timestamp
        >>> htlc: HTLC = HTLC(contract_address="xdcdE06b10c67765c8C0b9F64E0eF423b45Eb86b8e7", network="testnet")
        >>> htlc.build_htlc(secret_hash=sha256("Hello Meheret!"), recipient_address="xdcf8D43806260CFc6cC79fB408BA1897054667F81C", sender_address="xdc2224caA2235DF8Da3D2016d2AB1137D2d548A232", endtime=get_current_timestamp(plus=3600))
        >>> fund_transaction: FundTransaction = FundTransaction(network="testnet")
        >>> fund_transaction.build_transaction(address="xdc2224caA2235DF8Da3D2016d2AB1137D2d548A232", htlc=htlc, amount=3, unit="XDC")
        <swap.providers.xinfin.transaction.FundTransaction object at 0x0409DAF0>
        """

        # Check parameter instances
        if not is_address(address=address):
            raise AddressError(f"Invalid XinFin sender '{address}' address.")
        if not isinstance(htlc, HTLC):
            raise TypeError("Invalid XinFin HTLC instance, only takes XinFin HTLC class")
        if to_checksum_address(address=address, prefix="xdc") != htlc.agreements["sender_address"]:
            raise AddressError(f"Wrong XinFin sender '{address}' address",
                               "address must be match with HTLC agreements sender address.")
        if unit not in ["XDC", "Gwei", "Wei"]:
            raise UnitError("Invalid XinFin unit, choose only 'XDC', 'Gwei' or 'Wei' units.")

        _amount: Wei = Wei(
            amount if unit == "Wei" else amount_unit_converter(amount=amount, unit_from=f"{unit}2Wei")
        )

        htlc_contract: Contract = self.web3.eth.contract(
            address=self.web3.toChecksumAddress(htlc.contract_address(prefix="0x")), abi=htlc.abi()
        )

        htlc_fund_function = htlc_contract.functions.fund(
            unhexlify(htlc.agreements["secret_hash"]),  # Secret Hash
            to_checksum_address(htlc.agreements["recipient_address"], prefix="0x"),  # Recipient Address
            to_checksum_address(htlc.agreements["sender_address"], prefix="0x"),  # Sender Address
            htlc.agreements["endtime"]["timestamp"]  # Locktime Seconds
        )

        self._fee = htlc_fund_function.estimateGas({
            "from": to_checksum_address(address=address, prefix="0x"),
            "value": _amount,
            "nonce": self.web3.eth.get_transaction_count(
                to_checksum_address(address=address, prefix="0x")
            ),
            "gasPrice": self.web3.eth.gas_price
        })

        self._transaction = htlc_fund_function.buildTransaction({
            "from": to_checksum_address(address=address, prefix="0x"),
            "value": _amount,
            "nonce": self.web3.eth.get_transaction_count(
                to_checksum_address(address=address, prefix="0x")
            ),
            "gas": self._fee,
            "gasPrice": self.web3.eth.gas_price
        })
        self._type = "xinfin_fund_unsigned"
        return self

    def sign(self, solver: FundSolver) -> "FundTransaction":
        """
        Sign XinFin fund transaction.

        :param solver: XinFin fund solver.
        :type solver: xinfin.solver.FundSolver

        :returns: FundTransaction -- XinFin fund transaction instance.

        >>> from swap.providers.xinfin.htlc import HTLC
        >>> from swap.providers.xinfin.transaction import FundTransaction
        >>> from swap.providers.xinfin.solver import FundSolver
        >>> from swap.utils import sha256, get_current_timestamp
        >>> htlc: HTLC = HTLC(contract_address="xdcdE06b10c67765c8C0b9F64E0eF423b45Eb86b8e7", network="testnet")
        >>> htlc.build_htlc(secret_hash=sha256("Hello Meheret!"), recipient_address="xdcf8D43806260CFc6cC79fB408BA1897054667F81C", sender_address="xdc2224caA2235DF8Da3D2016d2AB1137D2d548A232", endtime=get_current_timestamp(plus=3600))
        >>> fund_transaction: FundTransaction = FundTransaction(network="testnet")
        >>> fund_transaction.build_transaction(address="xdc2224caA2235DF8Da3D2016d2AB1137D2d548A232", htlc=htlc, amount=3, unit="XDC")
        >>> fund_solver: FundSolver = FundSolver(xprivate_key="xprv9s21ZrQH143K3Y3pdbkbjreZQ9RVmqTLhRgf86uZyCJk2ou36YdUJt5frjwihGWmV1fQEDioiGZXWXUbHLy3kQf5xmhvhp8dZ2tfn6tgGUj", path="m/44'/550'/0'/0/0")
        >>> fund_transaction.sign(solver=fund_solver)
        <swap.providers.xinfin.transaction.FundTransaction object at 0x0409DAF0>
        """

        # Check parameter instances
        if not isinstance(solver, FundSolver):
            raise TypeError(f"Solver must be XinFin FundSolver, not {type(solver).__name__} type.")

        wallet: Wallet = solver.solve()
        signed_fund_transaction: SignedTransaction = self.web3.eth.account.sign_transaction(
            transaction_dict=self._transaction,
            private_key=wallet.private_key()
        )

        self._signature = dict(
            hash=signed_fund_transaction["hash"].hex(),
            rawTransaction=signed_fund_transaction["rawTransaction"].hex(),
            r=signed_fund_transaction["r"],
            s=signed_fund_transaction["s"],
            v=signed_fund_transaction["v"]
        )
        self._type = "xinfin_fund_signed"
        return self


class WithdrawTransaction(Transaction):
    """
    XinFin Withdraw transaction.

    :param network: XinFin network, defaults to ``mainnet``.
    :type network: str
    :param provider: XinFin network provider, defaults to ``http``.
    :type provider: str

    :returns: WithdrawTransaction -- XinFin withdraw transaction instance.

    .. warning::
        Do not forget to build transaction after initialize withdraw transaction.
    """

    def __init__(self, network: str = config["network"], provider: str = config["provider"]):
        super().__init__(
            network=network, provider=provider
        )

    def build_transaction(self, transaction_hash: str, address: str, secret_key: str,
                          contract_address: Optional[str] = None) -> "WithdrawTransaction":
        """
        Build XinFin withdraw transaction.

        :param transaction_hash: XinFin HTLC funded transaction hash.
        :type transaction_hash: str
        :param address: XinFin recipient address.
        :type address: str
        :param secret_key: Secret password/passphrase.
        :type secret_key: str
        :param contract_address: XinFin HTLC contract address, defaults to ``None``.
        :type contract_address: str

        :returns: WithdrawTransaction -- XinFin withdraw transaction instance.

        >>> from swap.providers.xinfin.transaction import WithdrawTransaction
        >>> withdraw_transaction: WithdrawTransaction = WithdrawTransaction(network="testnet")
        >>> withdraw_transaction.build_transaction(transaction_hash="0x0d4c93546aa3e5e476455931a63f1a97a2624e3b516e3fd8e3a582cb20aaeef9", secret_key="Hello Meheret!", address="xdcf8D43806260CFc6cC79fB408BA1897054667F81C", contract_address="xdcdE06b10c67765c8C0b9F64E0eF423b45Eb86b8e7")
        <swap.providers.xinfin.transaction.WithdrawTransaction object at 0x0409DAF0>
        """

        # Check parameter instances
        if not is_address(address=address):
            raise AddressError(f"Invalid XinFin recipient '{address}' address.")
        if contract_address and not is_address(address=contract_address):
            raise AddressError(f"Invalid XinFin HTLC contract '{contract_address}' address.")

        htlc: HTLC = HTLC(
            contract_address=contract_address, network=self._network
        )
        htlc_contract: Contract = self.web3.eth.contract(
            address=self.web3.toChecksumAddress(htlc.contract_address(prefix="0x")), abi=htlc.abi()
        )

        transaction_receipt: AttributeDict = _AttributeDict(get_transaction_receipt(
            transaction_hash=transaction_hash, network=self._network
        )).__attribute_dict__()
        log_fund: AttributeDict = htlc_contract.events.log_fund().processLog(
            log=transaction_receipt["logs"][0]
        )

        locked_contract_id: str = log_fund["args"]["locked_contract_id"]
        htlc_fund_function = htlc_contract.functions.withdraw(
            locked_contract_id,  # Locked Contract ID
            secret_key  # Secret Key
        )

        self._fee = htlc_fund_function.estimateGas({
            "from": to_checksum_address(address=address, prefix="0x"),
            "value": Wei(0),
            "nonce": self.web3.eth.get_transaction_count(
                to_checksum_address(address=address, prefix="0x")
            ),
            "gasPrice": self.web3.eth.gas_price
        })

        self._transaction = htlc_fund_function.buildTransaction({
            "from": to_checksum_address(address=address, prefix="0x"),
            "value": Wei(0),
            "nonce": self.web3.eth.get_transaction_count(
                to_checksum_address(address=address, prefix="0x")
            ),
            "gas": self._fee,
            "gasPrice": self.web3.eth.gas_price
        })
        self._type = "xinfin_withdraw_unsigned"
        return self

    def sign(self, solver: WithdrawSolver) -> "WithdrawTransaction":
        """
        Sign XinFin withdraw transaction.

        :param solver: XinFin withdraw solver.
        :type solver: xinfin.solver.WithdrawSolver

        :returns: WithdrawTransaction -- XinFin withdraw transaction instance.

        >>> from swap.providers.xinfin.transaction import WithdrawTransaction
        >>> from swap.providers.xinfin.solver import WithdrawSolver
        >>> withdraw_transaction: WithdrawTransaction = WithdrawTransaction(network="testnet")
        >>> withdraw_transaction.build_transaction(transaction_hash="0x0d4c93546aa3e5e476455931a63f1a97a2624e3b516e3fd8e3a582cb20aaeef9", secret_key="Hello Meheret!", address="xdcf8D43806260CFc6cC79fB408BA1897054667F81C", contract_address="xdcdE06b10c67765c8C0b9F64E0eF423b45Eb86b8e7")
        >>> withdraw_solver: WithdrawSolver = WithdrawSolver(xprivate_key="xprv9s21ZrQH143K4Kpce43z5guPyxLrFoc2i8aQAq835Zzp4Rt7i6nZaMCnVSDyHT6MnmJJGKHMrCUqaYpGojrug1ZN5qQDdShQffmkyv5xyUR", path="m/44'/550'/0'/0/0")
        >>> withdraw_transaction.sign(solver=withdraw_solver)
        <swap.providers.xinfin.transaction.WithdrawTransaction object at 0x0409DAF0>
        """

        # Check parameter instances
        if not isinstance(solver, WithdrawSolver):
            raise TypeError(f"Solver must be XinFin WithdrawSolver, not {type(solver).__name__} type.")

        wallet: Wallet = solver.solve()
        signed_withdraw_transaction: SignedTransaction = self.web3.eth.account.sign_transaction(
            transaction_dict=self._transaction,
            private_key=wallet.private_key()
        )

        self._signature = dict(
            hash=signed_withdraw_transaction["hash"].hex(),
            rawTransaction=signed_withdraw_transaction["rawTransaction"].hex(),
            r=signed_withdraw_transaction["r"],
            s=signed_withdraw_transaction["s"],
            v=signed_withdraw_transaction["v"]
        )
        self._type = "xinfin_withdraw_signed"
        return self


class RefundTransaction(Transaction):
    """
    XinFin Refund transaction.

    :param network: XinFin network, defaults to ``mainnet``.
    :type network: str
    :param provider: XinFin network provider, defaults to ``http``.
    :type provider: str

    :returns: RefundTransaction -- XinFin refund transaction instance.

    .. warning::
        Do not forget to build transaction after initialize refund transaction.
    """

    def __init__(self, network: str = config["network"], provider: str = config["provider"]):
        super().__init__(
            network=network, provider=provider
        )

    def build_transaction(self, transaction_hash: str, address: str,
                          contract_address: Optional[str] = None) -> "RefundTransaction":
        """
        Build XinFin refund transaction.

        :param transaction_hash: XinFin HTLC funded transaction hash.
        :type transaction_hash: str
        :param address: XinFin sender address.
        :type address: str
        :param contract_address: XinFin HTLC contract address, defaults to ``None``.
        :type contract_address: str

        :returns: RefundTransaction -- XinFin refund transaction instance.

        >>> from swap.providers.xinfin.transaction import RefundTransaction
        >>> refund_transaction: RefundTransaction = RefundTransaction(network="testnet")
        >>> refund_transaction.build_transaction(transaction_hash="0x0d4c93546aa3e5e476455931a63f1a97a2624e3b516e3fd8e3a582cb20aaeef9", address="xdc2224caA2235DF8Da3D2016d2AB1137D2d548A232", contract_address="xdcdE06b10c67765c8C0b9F64E0eF423b45Eb86b8e7")
        <swap.providers.xinfin.transaction.RefundTransaction object at 0x0409DAF0>
        """

        # Check parameter instances
        if not is_address(address=address):
            raise AddressError(f"Invalid XinFin sender '{address}' address.")
        if contract_address and not is_address(address=contract_address):
            raise AddressError(f"Invalid XinFin HTLC contract '{contract_address}' address.")

        htlc: HTLC = HTLC(
            contract_address=contract_address, network=self._network
        )
        htlc_contract: Contract = self.web3.eth.contract(
            address=self.web3.toChecksumAddress(htlc.contract_address(prefix="0x")), abi=htlc.abi()
        )

        transaction_receipt: AttributeDict = _AttributeDict(get_transaction_receipt(
            transaction_hash=transaction_hash, network=self._network
        )).__attribute_dict__()
        log_fund: AttributeDict = htlc_contract.events.log_fund().processLog(
            log=transaction_receipt["logs"][0]
        )

        locked_contract_id: str = log_fund["args"]["locked_contract_id"]
        htlc_refund_function = htlc_contract.functions.refund(
            locked_contract_id  # Locked Contract ID
        )

        self._fee = htlc_refund_function.estimateGas({
            "from": to_checksum_address(address=address, prefix="0x"),
            "value": Wei(0),
            "nonce": self.web3.eth.get_transaction_count(
                to_checksum_address(address=address, prefix="0x")
            ),
            "gasPrice": self.web3.eth.gas_price
        })

        self._transaction = htlc_refund_function.buildTransaction({
            "from": to_checksum_address(address=address, prefix="0x"),
            "value": Wei(0),
            "nonce": self.web3.eth.get_transaction_count(
                to_checksum_address(address=address, prefix="0x")
            ),
            "gas": self._fee,
            "gasPrice": self.web3.eth.gas_price
        })
        self._type = "xinfin_refund_unsigned"
        return self

    def sign(self, solver: RefundSolver) -> "RefundTransaction":
        """
        Sign XinFin refund transaction.

        :param solver: XinFin refund solver.
        :type solver: xinfin.solver.RefundSolver

        :returns: RefundTransaction -- XinFin refund transaction instance.

        >>> from swap.providers.xinfin.transaction import RefundTransaction
        >>> from swap.providers.xinfin.solver import RefundSolver
        >>> refund_transaction: RefundTransaction = RefundTransaction(network="testnet")
        >>> refund_transaction.build_transaction(transaction_hash="0x0d4c93546aa3e5e476455931a63f1a97a2624e3b516e3fd8e3a582cb20aaeef9", address="xdc2224caA2235DF8Da3D2016d2AB1137D2d548A232", contract_address="xdcdE06b10c67765c8C0b9F64E0eF423b45Eb86b8e7")
        >>> refund_solver: RefundSolver = RefundSolver(xprivate_key="xprv9s21ZrQH143K3Y3pdbkbjreZQ9RVmqTLhRgf86uZyCJk2ou36YdUJt5frjwihGWmV1fQEDioiGZXWXUbHLy3kQf5xmhvhp8dZ2tfn6tgGUj", path="m/44'/550'/0'/0/0")
        >>> refund_transaction.sign(solver=refund_solver)
        <swap.providers.xinfin.transaction.RefundTransaction object at 0x0409DAF0>
        """

        # Check parameter instances
        if not isinstance(solver, RefundSolver):
            raise TypeError(f"Solver must be XinFin RefundSolver, not {type(solver).__name__} type.")

        wallet: Wallet = solver.solve()
        signed_refund_transaction: SignedTransaction = self.web3.eth.account.sign_transaction(
            transaction_dict=self._transaction,
            private_key=wallet.private_key()
        )

        self._signature = dict(
            hash=signed_refund_transaction["hash"].hex(),
            rawTransaction=signed_refund_transaction["rawTransaction"].hex(),
            r=signed_refund_transaction["r"],
            s=signed_refund_transaction["s"],
            v=signed_refund_transaction["v"]
        )
        self._type = "xinfin_refund_signed"
        return self



#!/usr/bin/env python3

import json
import os

from swap.providers.vapor.htlc import HTLC
from swap.providers.vapor.transaction import (
    FundTransaction, WithdrawTransaction, RefundTransaction
)
from swap.providers.vapor.solver import (
    FundSolver, WithdrawSolver, RefundSolver
)
from swap.providers.vapor.utils import amount_unit_converter
from swap.utils import clean_transaction_raw

# Test Values
base_path = os.path.dirname(__file__)
file_path = os.path.abspath(os.path.join(base_path, "..", "..", "values.json"))
values = open(file_path, "r")
_ = json.loads(values.read())
values.close()


def test_vapor_fund_transaction():

    htlc = HTLC(network=_["vapor"]["network"]).build_htlc(
        secret_hash=_["vapor"]["htlc"]["secret"]["hash"],
        recipient_public_key=_["vapor"]["wallet"]["recipient"]["public_key"],
        sender_public_key=_["vapor"]["wallet"]["sender"]["public_key"],
        endblock=_["vapor"]["htlc"]["endblock"]
    )

    unsigned_fund_transaction = FundTransaction(network=_["vapor"]["network"])

    unsigned_fund_transaction.build_transaction(
        address=_["vapor"]["wallet"]["sender"]["address"],
        htlc=htlc,
        asset=_["vapor"]["asset"],
        amount=_["vapor"]["amount"],
        unit=_["vapor"]["unit"]
    )

    assert unsigned_fund_transaction.type() == _["vapor"]["fund"]["unsigned"]["type"]
    assert unsigned_fund_transaction.fee() == _["vapor"]["fund"]["unsigned"]["fee"]
    assert unsigned_fund_transaction.hash() == _["vapor"]["fund"]["unsigned"]["hash"]
    assert unsigned_fund_transaction.raw() == _["vapor"]["fund"]["unsigned"]["raw"]
    # assert unsigned_fund_transaction.json() == _["vapor"]["fund"]["unsigned"]["json"]
    assert unsigned_fund_transaction.unsigned_datas() == _["vapor"]["fund"]["unsigned"]["unsigned_datas"]
    assert unsigned_fund_transaction.signatures() == _["vapor"]["fund"]["unsigned"]["signatures"]
    assert unsigned_fund_transaction.transaction_raw() == clean_transaction_raw(
        transaction_raw=_["vapor"]["fund"]["unsigned"]["transaction_raw"]
    )

    signed_fund_transaction = unsigned_fund_transaction.sign(
        solver=FundSolver(
            xprivate_key=_["vapor"]["wallet"]["sender"]["xprivate_key"],
            path=_["vapor"]["wallet"]["sender"]["derivation"]["path"],
            account=_["vapor"]["wallet"]["sender"]["derivation"]["account"],
            change=_["vapor"]["wallet"]["sender"]["derivation"]["change"],
            address=_["vapor"]["wallet"]["sender"]["derivation"]["address"]
        )
    )

    assert signed_fund_transaction.type() == _["vapor"]["fund"]["signed"]["type"]
    assert signed_fund_transaction.fee() == _["vapor"]["fund"]["signed"]["fee"]
    assert signed_fund_transaction.hash() == _["vapor"]["fund"]["signed"]["hash"]
    assert signed_fund_transaction.raw() == _["vapor"]["fund"]["signed"]["raw"]
    # assert signed_fund_transaction.json() == _["vapor"]["fund"]["signed"]["json"]
    assert signed_fund_transaction.unsigned_datas() == _["vapor"]["fund"]["signed"]["unsigned_datas"]
    assert signed_fund_transaction.signatures() == _["vapor"]["fund"]["signed"]["signatures"]
    assert signed_fund_transaction.transaction_raw() == clean_transaction_raw(
        transaction_raw=_["vapor"]["fund"]["signed"]["transaction_raw"]
    )


def test_vapor_withdraw_transaction():

    unsigned_withdraw_transaction = WithdrawTransaction(network=_["vapor"]["network"])

    unsigned_withdraw_transaction.build_transaction(
        address=_["vapor"]["wallet"]["recipient"]["address"],
        transaction_hash=_["vapor"]["transaction_hash"],
        asset=_["vapor"]["asset"]
    )

    assert unsigned_withdraw_transaction.type() == _["vapor"]["withdraw"]["unsigned"]["type"]
    assert unsigned_withdraw_transaction.fee() == _["vapor"]["withdraw"]["unsigned"]["fee"]
    assert unsigned_withdraw_transaction.hash() == _["vapor"]["withdraw"]["unsigned"]["hash"]
    assert unsigned_withdraw_transaction.raw() == _["vapor"]["withdraw"]["unsigned"]["raw"]
    # assert unsigned_withdraw_transaction.json() == _["vapor"]["withdraw"]["unsigned"]["json"]
    assert unsigned_withdraw_transaction.unsigned_datas() == _["vapor"]["withdraw"]["unsigned"]["unsigned_datas"]
    assert unsigned_withdraw_transaction.signatures() == _["vapor"]["withdraw"]["unsigned"]["signatures"]
    assert unsigned_withdraw_transaction.transaction_raw() == clean_transaction_raw(
        transaction_raw=_["vapor"]["withdraw"]["unsigned"]["transaction_raw"]
    )

    signed_withdraw_transaction = unsigned_withdraw_transaction.sign(
        solver=WithdrawSolver(
            xprivate_key=_["vapor"]["wallet"]["recipient"]["xprivate_key"],
            secret_key=_["vapor"]["htlc"]["secret"]["key"],
            bytecode=_["vapor"]["htlc"]["bytecode"],
            path=_["vapor"]["wallet"]["recipient"]["derivation"]["path"],
            account=_["vapor"]["wallet"]["recipient"]["derivation"]["account"],
            change=_["vapor"]["wallet"]["recipient"]["derivation"]["change"],
            address=_["vapor"]["wallet"]["recipient"]["derivation"]["address"]
        )
    )

    assert signed_withdraw_transaction.type() == _["vapor"]["withdraw"]["signed"]["type"]
    assert signed_withdraw_transaction.fee() == _["vapor"]["withdraw"]["signed"]["fee"]
    assert signed_withdraw_transaction.hash() == _["vapor"]["withdraw"]["signed"]["hash"]
    assert signed_withdraw_transaction.raw() == _["vapor"]["withdraw"]["signed"]["raw"]
    # assert signed_withdraw_transaction.json() == _["vapor"]["withdraw"]["signed"]["json"]
    assert signed_withdraw_transaction.unsigned_datas() == _["vapor"]["withdraw"]["signed"]["unsigned_datas"]
    assert signed_withdraw_transaction.signatures() == _["vapor"]["withdraw"]["signed"]["signatures"]
    assert signed_withdraw_transaction.transaction_raw() == clean_transaction_raw(
        transaction_raw=_["vapor"]["withdraw"]["signed"]["transaction_raw"]
    )


def test_vapor_refund_transaction():

    unsigned_refund_transaction = RefundTransaction(network=_["vapor"]["network"])

    unsigned_refund_transaction.build_transaction(
        address=_["vapor"]["wallet"]["sender"]["address"],
        transaction_hash=_["vapor"]["transaction_hash"],
        asset=_["vapor"]["asset"]
    )

    assert unsigned_refund_transaction.type() == _["vapor"]["refund"]["unsigned"]["type"]
    assert unsigned_refund_transaction.fee() == _["vapor"]["refund"]["unsigned"]["fee"]
    assert unsigned_refund_transaction.hash() == _["vapor"]["refund"]["unsigned"]["hash"]
    assert unsigned_refund_transaction.raw() == _["vapor"]["refund"]["unsigned"]["raw"]
    # assert unsigned_refund_transaction.json() == _["vapor"]["refund"]["unsigned"]["json"]
    assert unsigned_refund_transaction.unsigned_datas() == _["vapor"]["refund"]["unsigned"]["unsigned_datas"]
    assert unsigned_refund_transaction.signatures() == _["vapor"]["refund"]["unsigned"]["signatures"]
    assert unsigned_refund_transaction.transaction_raw() == clean_transaction_raw(
        transaction_raw=_["vapor"]["refund"]["unsigned"]["transaction_raw"]
    )

    signed_refund_transaction = unsigned_refund_transaction.sign(
        solver=RefundSolver(
            xprivate_key=_["vapor"]["wallet"]["sender"]["xprivate_key"],
            bytecode=_["vapor"]["htlc"]["bytecode"],
            path=_["vapor"]["wallet"]["sender"]["derivation"]["path"],
            account=_["vapor"]["wallet"]["sender"]["derivation"]["account"],
            change=_["vapor"]["wallet"]["sender"]["derivation"]["change"],
            address=_["vapor"]["wallet"]["sender"]["derivation"]["address"]
        )
    )

    assert signed_refund_transaction.type() == _["vapor"]["refund"]["signed"]["type"]
    assert signed_refund_transaction.fee() == _["vapor"]["refund"]["signed"]["fee"]
    assert signed_refund_transaction.hash() == _["vapor"]["refund"]["signed"]["hash"]
    assert signed_refund_transaction.raw() == _["vapor"]["refund"]["signed"]["raw"]
    # assert signed_refund_transaction.json() == _["vapor"]["refund"]["signed"]["json"]
    assert signed_refund_transaction.unsigned_datas() == _["vapor"]["refund"]["signed"]["unsigned_datas"]
    assert signed_refund_transaction.signatures() == _["vapor"]["refund"]["signed"]["signatures"]
    assert signed_refund_transaction.transaction_raw() == clean_transaction_raw(
        transaction_raw=_["vapor"]["refund"]["signed"]["transaction_raw"]
    )



#!/usr/bin/env python3

from typing import (
    Optional, Union
)

import requests
import json

from ...exceptions import (
    BalanceError, APIError, NetworkError, AddressError
)
from ..config import bytom as config
from .assets import AssetNamespace
from .utils import (
    is_network, is_address, amount_unit_converter, get_address_type
)


def get_balance(address: str, asset: Union[str, AssetNamespace] = config["asset"], network: str = config["network"],
                headers: dict = config["headers"], timeout: int = config["timeout"]) -> int:
    """
    Get Bytom balance.

    :param address: Bytom address.
    :type address: str
    :param asset: Bytom asset, default to ``BTM``.
    :type asset: str, bytom.assets.AssetNamespace
    :param network: Bytom network, defaults to ``mainnet``.
    :type network: str
    :param headers: Request headers, default to ``common headers``.
    :type headers: dict
    :param timeout: Request timeout, default to ``60``.
    :type timeout: int

    :returns: int -- Bytom asset balance (NEU amount).

    >>> from swap.providers.bytom.rpc import get_balance
    >>> from swap.providers.bytom.assets import BTM as ASSET
    >>> get_balance(address="bm1qk9vj4jaezlcnjdckds4fkm8fwv5kawmq9qrufx", asset=ASSET, network="mainnet")
    71560900
    """

    if not is_address(address=address, network=network):
        raise AddressError(f"Invalid Bytom '{address}' {network} address.")
    if not is_network(network=network):
        raise NetworkError(f"Invalid Bytom '{network}' network",
                           "choose only 'mainnet', 'solonet' or 'testnet' networks.")

    url = f"{config[network]['blockmeta']}/address/{address}/asset"
    response = requests.get(
        url=url, headers=headers, timeout=timeout
    )
    response_json = response.json()
    if response_json is None:
        return 0
    for _asset in response_json:
        if (str(asset.ID) if isinstance(asset, AssetNamespace) else asset) == _asset["asset_id"]:
            return int(_asset["balance"])
    return 0


def get_utxos(program: str, network: str = config["network"], asset: Union[str, AssetNamespace] = config["asset"],
              limit: int = 15, by: str = "amount", order: str = "desc",
              headers: dict = config["headers"], timeout: int = config["timeout"]) -> list:
    """
    Get Bytom unspent transaction outputs (UTXO's).

    :param program: Bytom control program.
    :type program: str
    :param network: Bytom network, defaults to ``mainnet``.
    :type network: str
    :param asset: Bytom asset id, defaults to ``BTM``.
    :type asset: str, bytom.assets.AssetNamespace
    :param limit: Bytom utxo's limit, defaults to ``15``.
    :type limit: int
    :param by: Sort by, defaults to ``amount``.
    :type by: str
    :param order: Sort order, defaults to ``desc``.
    :type order: str
    :param headers: Request headers, default to ``common headers``.
    :type headers: dict
    :param timeout: Request timeout, default to ``60``.
    :type timeout: int

    :returns: list -- Bytom unspent transaction outputs (UTXO's).

    >>> from swap.providers.bytom.rpc import get_utxos
    >>> get_utxos(program="00142cda4f99ea8112e6fa61cdd26157ed6dc408332a", network="mainnet")
    [{'hash': '7c1e20e6ff719176a3ed6f5332ec3ff665ab28754d2511950e591267e0e675df', 'asset': 'ffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffff', 'amount': 71510800}, {'hash': '01b07c3523085b75f1e047be3a73b263635d0b86f9b751457a51b26c5a97a110', 'asset': 'ffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffff', 'amount': 50000}, {'hash': 'e46cfecc1f1a26413172ce81c78affb19408e613915642fa5fb04d3b0a4ffa65', 'asset': 'ffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffff', 'amount': 100}]
    """

    if not is_network(network=network):
        raise NetworkError(f"Invalid Bytom '{network}' network",
                           "choose only 'mainnet' or 'testnet' networks.")

    url = f"{config[network]['blockcenter']}/q/utxos"
    data = dict(filter=dict(
        script=program, asset=(str(asset.ID) if isinstance(asset, AssetNamespace) else asset)
    ), sort=dict(by=by, order=order))
    params = dict(limit=limit)
    response = requests.post(
        url=url, data=json.dumps(data), params=params, headers=headers, timeout=timeout
    )
    response_json = response.json()
    return response_json["data"]


def estimate_transaction_fee(address: str, amount: int, asset: Union[str, AssetNamespace] = config["asset"],
                             confirmations: int = config["confirmations"], network: str = config["network"],
                             headers: dict = config["headers"], timeout: int = config["timeout"]) -> int:
    """
    Estimate Bytom transaction fee.

    :param address: Bytom address.
    :type address: str
    :param amount: Bytom amount (NEU amount).
    :type amount: int
    :param asset: Bytom asset id, default to ``BTM``.
    :type asset: str, bytom.assets.AssetNamespace
    :param confirmations: Bytom confirmations, default to ``1``.
    :type confirmations: int
    :param network: Bytom network, defaults to ``mainnet``.
    :type network: str
    :param headers: Request headers, default to ``common headers``.
    :type headers: dict
    :param timeout: request timeout, default to ``60``.
    :type timeout: int

    :returns: str -- Estimated transaction fee (NEU amount).

    >>> from swap.providers.bytom.rpc import estimate_transaction_fee
    >>> from swap.providers.bytom.assets import BTM as ASSET
    >>> estimate_transaction_fee(address="bm1qk9vj4jaezlcnjdckds4fkm8fwv5kawmq9qrufx", asset=ASSET, amount=100_000, confirmations=100, network="mainnet")
    449000
    """

    if not is_network(network=network):
        raise NetworkError(f"Invalid '{network}' network",
                           "choose only 'mainnet', 'solonet' or 'testnet' networks.")
    if not is_address(address=address, network=network):
        raise AddressError(f"Invalid Bytom '{address}' {network} address.")

    url = f"{config[network]['blockcenter']}/merchant/estimate-tx-fee"
    data = dict(
        asset_amounts={
            (str(asset.ID) if isinstance(asset, AssetNamespace) else asset): str(amount_unit_converter(
                amount=amount, unit_from="NEU2BTM"
            ))
        },
        confirmations=confirmations
    )
    params = dict(address=address)
    response = requests.post(
        url=url, data=json.dumps(data), params=params, headers=headers, timeout=timeout
    )
    if response.status_code == 200 and response.json()["code"] == 200:
        return amount_unit_converter(amount=float(response.json()["data"]["fee"]), unit_from="BTM2NEU")
    elif response.status_code == 200 and response.json()["code"] == 504:
        raise BalanceError("Insufficient spend UTXO's", "you don't have enough amount.")
    raise APIError(response.json()["msg"], response.json()["code"])


def account_create(xpublic_key: str, label: str = "1st address", account_index: int = 1,
                   network: str = config["network"], headers: dict = config["headers"],
                   timeout: int = config["timeout"]) -> dict:
    """
    Create account in blockcenter.

    :param xpublic_key: Bytom xpublic key.
    :type xpublic_key: str
    :param label: Bytom limit, defaults to ``1st address``.
    :type label: str
    :param account_index: Account index, defaults to ``1``.
    :type account_index: str
    :param network: Bytom network, defaults to ``mainnet``.
    :type network: str
    :param headers: Request headers, default to ``common headers``.
    :type headers: dict
    :param timeout: request timeout, default to ``60``.
    :type timeout: int

    :returns: dict -- Bytom blockcenter guid, address and label.

    >>> from swap.providers.bytom.rpc import account_create
    >>> account_create(xpublic_key="f80a401807fde1ee5727ae032ee144e4b757e69431e68e6cd732eda3c8cd3936daedfdd0fd8f8df14e2084c7e8df4701db3062dded1c713e0aae734ac09c4afd", network="mainnet")
    {"guid": "9ed61a9b-e7b6-4cb7-94fb-932b738e4f66", "address": "bm1qk9vj4jaezlcnjdckds4fkm8fwv5kawmq9qrufx", "label": "1st address"}
    """

    if not is_network(network=network):
        raise NetworkError(f"Invalid Bytom '{network}' network",
                           "choose only 'mainnet', 'solonet' or 'testnet' networks.")

    url = f"{config[network]['blockcenter']}/account/create"
    data = dict(pubkey=xpublic_key, label=label, account_index=account_index)
    response = requests.post(
        url=url, data=json.dumps(data), headers=headers, timeout=timeout
    )
    if response.status_code == 200 and response.json()["code"] == 200:
        return response.json()["data"]
    raise APIError(response.json()["msg"], response.json()["code"])


def build_transaction(address: str, transaction: dict, network: str = config["network"],
                      headers: dict = config["headers"], timeout: int = config["timeout"]) -> dict:
    """
    Build Bytom transaction.

    :param address: Bytom address.
    :type address: str
    :param transaction: Bytom transaction (inputs, outputs, fee, confirmations & forbid_chain_tx).
    :type transaction: dict
    :param network: Bytom network, defaults to ``mainnet``.
    :type network: str
    :param headers: Request headers, default to ``common headers``.
    :type headers: dict
    :param timeout: Request timeout, default to ``60``.
    :type timeout: int

    :returns: dict -- Bytom builted transaction.

    >>> from swap.providers.bytom.rpc import build_transaction
    >>> build_transaction(address="bm1qk9vj4jaezlcnjdckds4fkm8fwv5kawmq9qrufx", transaction={"fee": "0.1", "confirmations": 1, "inputs": [{"type": "spend_wallet", "amount": "0.0001", "asset": "ffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffff"}], "outputs": [{"type": "control_address", "amount": "0.0001", "asset": "ffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffff", "address": "bm1qf78sazxs539nmzztq7md63fk2x8lew6ed2gu5rnt9um7jerrh07q3yf5q8"}]}, network="mainnet")
    {'tx': {'hash': '5d4ae68487953863783599045f99eb8740b5745376ed8d8926d68de695e72476', 'status': True, 'size': 404, 'submission_timestamp': 0, 'memo': '', 'inputs': [{'script': '00142cda4f99ea8112e6fa61cdd26157ed6dc408332a', 'address': 'bm1qk9vj4jaezlcnjdckds4fkm8fwv5kawmq9qrufx', 'asset': {'asset_id': 'ffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffff', 'decimals': 0, 'unit': 'BTM'}, 'amount': '0.0005', 'type': 'spend'}, {'script': '00142cda4f99ea8112e6fa61cdd26157ed6dc408332a', 'address': 'bm1qk9vj4jaezlcnjdckds4fkm8fwv5kawmq9qrufx', 'asset': {'asset_id': 'ffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffff', 'decimals': 0, 'unit': 'BTM'}, 'amount': '0.715108', 'type': 'spend'}], 'outputs': [{'utxo_id': '0d5c097b8e75f711765ff63017fe8a4a987d8b50f7ca3a5d1873120af5f46116', 'script': '00204f8f0e88d0a44b3d884b07b6dd4536518ffcbb596a91ca0e6b2f37e96463bbfc', 'address': 'bm1qf78sazxs539nmzztq7md63fk2x8lew6ed2gu5rnt9um7jerrh07q3yf5q8', 'asset': {'asset_id': 'ffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffff', 'decimals': 0, 'unit': 'BTM'}, 'amount': '0.0001', 'type': 'control'}, {'utxo_id': 'c49da44ef15d227ca978191e91d5d8915a3f92baf6b5778b7377deb2bddca554', 'script': '00142cda4f99ea8112e6fa61cdd26157ed6dc408332a', 'address': 'bm1qk9vj4jaezlcnjdckds4fkm8fwv5kawmq9qrufx', 'asset': {'asset_id': 'ffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffff', 'decimals': 0, 'unit': 'BTM'}, 'amount': '0.615908', 'type': 'control'}], 'fee': '0.0996', 'balances': [{'asset': {'asset_id': 'ffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffff', 'decimals': 0, 'unit': 'BTM'}, 'amount': '-0.0001'}], 'types': ['ordinary'], 'min_veto_height': 0}, 'raw_transaction': '07010002015e015c88650475abf87eb364f93c608db879ad71643fbc7725ded246e8883e79c75a78ffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffd0860300011600142cda4f99ea8112e6fa61cdd26157ed6dc408332a22012091ff7f525ff40874c4f47f0cab42e46e3bf53adad59adef9558ad1b6448f22e2015f015da72ea4ad87d7b5a51534c07edc005887345ef38fac8d258987dd17268e8d0336ffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffff90d68c2201011600142cda4f99ea8112e6fa61cdd26157ed6dc408332a22012091ff7f525ff40874c4f47f0cab42e46e3bf53adad59adef9558ad1b6448f22e2020146ffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffff904e012200204f8f0e88d0a44b3d884b07b6dd4536518ffcbb596a91ca0e6b2f37e96463bbfc00013cffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffff909aaf1d011600142cda4f99ea8112e6fa61cdd26157ed6dc408332a00', 'signing_instructions': [{'derivation_path': ['2c000000', '99000000', '01000000', '00000000', '01000000'], 'sign_data': ['a5da2ae06bfaea9854423fe9cc544d775854cf57827c8c2ab606418452d30209'], 'pubkey': '91ff7f525ff40874c4f47f0cab42e46e3bf53adad59adef9558ad1b6448f22e2'}, {'derivation_path': ['2c000000', '99000000', '01000000', '00000000', '01000000'], 'sign_data': ['3e44203712c4e981783810875fa67f2efe0afda38afe229fd09da0d113c3d885'], 'pubkey': '91ff7f525ff40874c4f47f0cab42e46e3bf53adad59adef9558ad1b6448f22e2'}]}
    """

    if not is_address(address=address, network=network):
        raise AddressError(f"Invalid Bytom '{address}' {network} address.")
    if not is_network(network=network):
        raise NetworkError(f"Invalid Bytom '{network}' network",
                           "choose only 'mainnet', 'solonet' or 'testnet' networks.")

    url = f"{config[network]['blockcenter']}/merchant/build-advanced-tx"
    params = dict(address=address)
    response = requests.post(
        url=url, data=json.dumps(transaction), params=params, headers=headers, timeout=timeout
    )
    if response.status_code == 200 and response.json()["code"] == 300:
        raise APIError(response.json()["msg"], response.json()["code"])
    elif response.status_code == 200 and response.json()["code"] == 503:
        raise APIError(response.json()["msg"], response.json()["code"])
    elif response.status_code == 200 and response.json()["code"] == 422:
        raise BalanceError(f"There is no any asset balance recorded on this '{address}' address.")
    elif response.status_code == 200 and response.json()["code"] == 515:
        raise BalanceError(f"Insufficient balance, check your balance and try again.")
    elif response.status_code == 200 and response.json()["code"] == 504:
        raise BalanceError(f"Insufficient balance, check your balance and try again.")
    return response.json()["data"][0]


def get_transaction(transaction_hash: str, network: str = config["network"],
                    headers: dict = config["headers"], timeout: int = config["timeout"]) -> dict:
    """
    Get Bytom transaction detail.

    :param transaction_hash: Bytom transaction hash/id.
    :type transaction_hash: str
    :param network: Bytom network, defaults to ``mainnet``.
    :type network: str
    :param headers: Request headers, default to ``common headers``.
    :type headers: dict
    :param timeout: Request timeout, default to ``60``.
    :type timeout: int

    :returns: dict -- Bytom transaction detail.

    >>> from swap.providers.bytom.rpc import get_transaction
    >>> get_transaction(transaction_hash="bc935995cb3408b51aa3d05e7e77226840eb68340b229f9c561edae31ebc8b95", network="mainnet")
    {'id': 'bc935995cb3408b51aa3d05e7e77226840eb68340b229f9c561edae31ebc8b95', 'timestamp': 1524765978, 'block_height': 3487, 'trx_amount': 41249562600, 'trx_fee': 437400, 'status_fail': False, 'coinbase': False, 'size': 332, 'chain_status': 'mainnet', 'time_range': 0, 'index_id': 3489, 'mux_id': '305a28d8d34b40c65936810f9e9c1f8bc9c793ec2e72c70f9203fbbeb0a56db9', 'inputs': [{'txtype': 'spend', 'asset_id': 'ffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffff', 'amount': 41250000000, 'control_program': '0014151df3db084d909ccb55d45d4e59db2e17e5f237', 'address': 'bm1qz5wl8kcgfkgfej6463w5ukwm9ct7tu3ht8p7te', 'spent_output_id': '6def8e6a7c29ccff4c5596a37a6698b71f392bf713bc67bb3fa0af54bf50f815', 'input_id': 'fb226e3ad39e38341f0d232c910065b76ef7c267faa3ea4e49a31836405b6747', 'witness_arguments': ['1848cb550620b971fd244eb625ccf4507ccd9944da65b47674550397c983247e1bd3ff880782beca963a81c34c17c8ef664e2501a11cdd9097300e44567ff10f', '268402537b02d91fafdcdeb6eda3aa542548d77cd6cccb38ecd7ea7ce8a22cf7'], 'asset_name': 'BTM', 'asset_definition': '{}', 'cross_chain_asset': False, 'asset_decimals': 8}], 'outputs': [{'txtype': 'control', 'id': 'a8a7b5363379dee8ff77da7c4acf63dc3469a79ebaade079fc1842543964c6e9', 'asset_id': 'ffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffff', 'amount': 41239562600, 'control_program': '0014fc22634a713ac1e6f831c56184f847b7546fbda4', 'address': 'bm1qls3xxjn38tq7d7p3c4scf7z8ka2xl0dyppj52k', 'asset_name': 'BTM', 'asset_definition': '{}', 'cross_chain_asset': False, 'position': 0, 'asset_decimals': 8}, {'txtype': 'control', 'id': '84287fb5b2b461dbd3b937a9013d89c0d54a21768e31fb8345b02d57a7992533', 'asset_id': 'ffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffff', 'amount': 10000000, 'control_program': '00140e43a92a9e8aca788eb1551c316448c2e3f78215', 'address': 'bm1qpep6j2573t983r4325wrzezgct3l0qs4q04pem', 'asset_name': 'BTM', 'asset_definition': '{}', 'cross_chain_asset': False, 'position': 1, 'asset_decimals': 8}], 'confirmations': 558509}
    """

    if not is_network(network=network):
        raise NetworkError(f"Invalid Bytom '{network}' network",
                           "choose only 'mainnet', 'solonet' or 'testnet' networks.")

    url = f"{config[network]['blockmeta']}/transaction/{transaction_hash}"
    response = requests.get(
        url=url, headers=headers, timeout=timeout
    )
    if response.status_code == 200 and response.json()["inputs"] is not None:
        return response.json()
    raise APIError(f"Not found this '{transaction_hash}' transaction hash.", 500)


def get_current_block_height(plus: int = 0, network: str = config["network"],
                             headers: dict = config["headers"], timeout: int = config["timeout"]) -> int:
    """
    Get Bytom transaction detail.

    :param plus: Add block number on current block height, default to ``0``.
    :type plus: int
    :param network: Bytom network, defaults to ``mainnet``.
    :type network: str
    :param headers: Request headers, default to ``common headers``.
    :type headers: dict
    :param timeout: Request timeout, default to ``60``.
    :type timeout: int

    :returns: int -- Bytom current block height.

    >>> from swap.providers.bytom.rpc import get_current_block_height
    >>> get_current_block_height(plus=0)
    678722
    """

    if not is_network(network=network):
        raise NetworkError(f"Invalid Bytom '{network}' network",
                           "choose only 'mainnet', 'solonet' or 'testnet' networks.")

    url = f"{config[network]['blockmeta']}/latest-block"
    response = requests.get(
        url=url, headers=headers, timeout=timeout
    )
    if response.status_code == 200:
        return int(response.json()["block"]["height"]) \
            if plus == 0 else int(response.json()["block"]["height"]) + plus
    raise APIError("Can't get current latest Bytom block height.")


def find_p2wsh_utxo(transaction: dict) -> Optional[dict]:
    """
    Find Bytom pay to witness script hash UTXO info's.

    :param transaction: Bytom transaction detail.
    :type transaction: dict

    :returns: dict -- Pay to Witness Secript Hash (P2WSH) UTXO info's.

    >>> from swap.providers.bytom.rpc import find_p2wsh_utxo, get_transaction
    >>> find_p2wsh_utxo(transaction=get_transaction(transaction_hash="b6d12407bbd238938941246fd0dd3e5234f1e3c370bef3fcbc1f60ebee022e76", network="mainnet"))
    {'txtype': 'control', 'id': 'a1c5cce9df9343a10dafa582dea04e61c402ee8398b5268ba5c9c3aefd58017a', 'asset_id': 'ffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffff', 'amount': 10499000, 'control_program': '00204f8f0e88d0a44b3d884b07b6dd4536518ffcbb596a91ca0e6b2f37e96463bbfc', 'address': 'bm1qf78sazxs539nmzztq7md63fk2x8lew6ed2gu5rnt9um7jerrh07q3yf5q8', 'asset_name': 'BTM', 'asset_definition': '{}', 'cross_chain_asset': False, 'position': 0, 'asset_decimals': 8}
    """

    transaction_outputs, utxo = transaction["outputs"], None
    for transaction_output in transaction_outputs:
        if get_address_type(transaction_output["address"]) == "p2wsh":
            utxo = transaction_output
            break
    return utxo


def decode_raw(raw: str, network: str = config["network"], 
               headers: dict = config["headers"], timeout: int = config["timeout"]) -> dict:
    """
    Decode original Bytom raw.

    :param raw: Bytom transaction raw.
    :type raw: str
    :param network: Bytom network, defaults to ``mainnet``.
    :type network: str
    :param headers: Request headers, default to ``common headers``.
    :type headers: dict
    :param timeout: Request timeout, default to ``60``.
    :type timeout: int

    :returns: dict -- Bytom decoded transaction raw.

    >>> from swap.providers.bytom.rpc import decode_raw
    >>> decode_raw(raw="07010002015e015c88650475abf87eb364f93c608db879ad71643fbc7725ded246e8883e79c75a78ffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffd0860300011600142cda4f99ea8112e6fa61cdd26157ed6dc408332a22012091ff7f525ff40874c4f47f0cab42e46e3bf53adad59adef9558ad1b6448f22e2015f015da72ea4ad87d7b5a51534c07edc005887345ef38fac8d258987dd17268e8d0336ffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffff90d68c2201011600142cda4f99ea8112e6fa61cdd26157ed6dc408332a22012091ff7f525ff40874c4f47f0cab42e46e3bf53adad59adef9558ad1b6448f22e2020146ffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffff904e012200204f8f0e88d0a44b3d884b07b6dd4536518ffcbb596a91ca0e6b2f37e96463bbfc00013cffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffff909aaf1d011600142cda4f99ea8112e6fa61cdd26157ed6dc408332a00", network="testnet")
    {'tx_id': '5d4ae68487953863783599045f99eb8740b5745376ed8d8926d68de695e72476', 'version': 1, 'size': 404, 'time_range': 0, 'inputs': [{'type': 'spend', 'asset_id': 'ffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffff', 'asset_definition': {}, 'amount': 50000, 'control_program': '00142cda4f99ea8112e6fa61cdd26157ed6dc408332a', 'address': 'bm1qk9vj4jaezlcnjdckds4fkm8fwv5kawmq9qrufx', 'spent_output_id': '01b07c3523085b75f1e047be3a73b263635d0b86f9b751457a51b26c5a97a110', 'input_id': 'de193c78772c93356f81a5061a90d8dcfba84d03ae4d78b2a57a9201f88c38af', 'witness_arguments': ['91ff7f525ff40874c4f47f0cab42e46e3bf53adad59adef9558ad1b6448f22e2'], 'sign_data': 'a5da2ae06bfaea9854423fe9cc544d775854cf57827c8c2ab606418452d30209'}, {'type': 'spend', 'asset_id': 'ffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffff', 'asset_definition': {}, 'amount': 71510800, 'control_program': '00142cda4f99ea8112e6fa61cdd26157ed6dc408332a', 'address': 'bm1qk9vj4jaezlcnjdckds4fkm8fwv5kawmq9qrufx', 'spent_output_id': '7c1e20e6ff719176a3ed6f5332ec3ff665ab28754d2511950e591267e0e675df', 'input_id': 'de2c7bcf9caf00f78ca8e316cf37cf88c86b0457e47cf58e2465d783151abd0e', 'witness_arguments': ['91ff7f525ff40874c4f47f0cab42e46e3bf53adad59adef9558ad1b6448f22e2'], 'sign_data': '3e44203712c4e981783810875fa67f2efe0afda38afe229fd09da0d113c3d885'}], 'outputs': [{'type': 'control', 'id': '0d5c097b8e75f711765ff63017fe8a4a987d8b50f7ca3a5d1873120af5f46116', 'position': 0, 'asset_id': 'ffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffff', 'asset_definition': {}, 'amount': 10000, 'control_program': '00204f8f0e88d0a44b3d884b07b6dd4536518ffcbb596a91ca0e6b2f37e96463bbfc', 'address': 'bm1qf78sazxs539nmzztq7md63fk2x8lew6ed2gu5rnt9um7jerrh07q3yf5q8'}, {'type': 'control', 'id': 'c49da44ef15d227ca978191e91d5d8915a3f92baf6b5778b7377deb2bddca554', 'position': 1, 'asset_id': 'ffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffff', 'asset_definition': {}, 'amount': 61590800, 'control_program': '00142cda4f99ea8112e6fa61cdd26157ed6dc408332a', 'address': 'bm1qk9vj4jaezlcnjdckds4fkm8fwv5kawmq9qrufx'}], 'fee': 9960000}
    """

    if not is_network(network=network):
        raise NetworkError(f"Invalid Bytom '{network}' network",
                           "choose only 'mainnet', 'solonet' or 'testnet' networks.")

    url = f"{config[network]['bytom-core']}/decode-raw-transaction"
    data = dict(raw_transaction=raw)
    response = requests.post(
        url=url, data=json.dumps(data), headers=headers, timeout=timeout
    )
    response_json = response.json()
    if response.status_code == 400:
        raise APIError(response_json["msg"], response_json["code"])
    return response_json["data"]


def submit_raw(address: str, raw: str, signatures: list, network: str = config["network"],
               headers: dict = config["headers"], timeout: int = config["timeout"]) -> str:
    """
     Submit original Bytom raw into blockchain.

    :param address: Bytom address.
    :type address: str
    :param raw: Bytom transaction raw.
    :type raw: str
    :param signatures: Bytom signed massage datas.
    :type signatures: list
    :param network: Bytom network, defaults to ``mainnet``.
    :type network: str
    :param headers: Request headers, default to ``common headers``.
    :type headers: dict
    :param timeout: Request timeout, default to ``60``.
    :type timeout: int

    :returns: str -- Bytom submitted transaction id/hash.

    >>> from swap.providers.bytom.rpc import submit_raw
    >>> submit_raw(address="bm1qk9vj4jaezlcnjdckds4fkm8fwv5kawmq9qrufx", raw="07010002015e015c88650475abf87eb364f93c608db879ad71643fbc7725ded246e8883e79c75a78ffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffd0860300011600142cda4f99ea8112e6fa61cdd26157ed6dc408332a22012091ff7f525ff40874c4f47f0cab42e46e3bf53adad59adef9558ad1b6448f22e2015f015da72ea4ad87d7b5a51534c07edc005887345ef38fac8d258987dd17268e8d0336ffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffff90d68c2201011600142cda4f99ea8112e6fa61cdd26157ed6dc408332a22012091ff7f525ff40874c4f47f0cab42e46e3bf53adad59adef9558ad1b6448f22e2020146ffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffff904e012200204f8f0e88d0a44b3d884b07b6dd4536518ffcbb596a91ca0e6b2f37e96463bbfc00013cffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffff909aaf1d011600142cda4f99ea8112e6fa61cdd26157ed6dc408332a00", signatures=[["f8466336a79d166e47fb5d64f1e7ec01b203b59b3ee86686492bd1e4d0bdd642dfe4a575049071a052a441635c336708ab7d869cccd5331bc29f60e0ed9cd80d"], ["ebf33fbda5c2f3d144e90c3b763b1e7e42d501e595216fcd2b310b089918bae2ef4c7b8a2e1f650ee741578aba7960706d2bf9be7dffbf0fe77199075f155909"]], network="mainnet")
    "2993414225f65390220730d0c1a356c14e91bca76db112d37366df93e364a492"
    """

    if not is_address(address=address, network=network):
        raise AddressError(f"Invalid Bytom '{address}' {network} address.")
    if not is_network(network=network):
        raise NetworkError(f"Invalid Bytom '{network}' network",
                           "choose only 'mainnet', 'solonet' or 'testnet' networks.")

    url = f"{config[network]['blockcenter']['v3']}/merchant/submit-payment"
    data = dict(raw_transaction=raw, signatures=signatures)
    params = dict(address=address)
    response = requests.post(
        url=url, data=json.dumps(data), params=params, headers=headers, timeout=timeout
    )
    response_json = response.json()
    if response_json["code"] != 200:
        raise APIError(response_json["msg"], response_json["code"])
    return response_json["data"]["tx_hash"]



#!/usr/bin/env python3

from base64 import b64decode
from datetime import datetime
from web3.types import ChecksumAddress
from hexbytes.main import HexBytes
from web3 import Web3
from typing import (
    Union, Optional
)

import json

from ...utils import clean_transaction_raw
from ...exceptions import (
    AddressError, UnitError, TransactionRawError
)
from ..config import ethereum as config


def is_network(network: str) -> bool:
    """
    Check Ethereum network.

    :param network: Ethereum network.
    :type network: str

    :returns: bool -- Ethereum valid/invalid network.

    >>> from swap.providers.ethereum.utils import is_network
    >>> is_network(network="kovan")
    True
    """

    # Check parameter instances
    if not isinstance(network, str):
        raise TypeError(f"Network must be str, not '{type(network)}' type.")

    return network in ["mainnet", "ropsten", "kovan", "rinkeby", "testnet"]


def is_address(address: str) -> bool:
    """
    Check Ethereum address.

    :param address: Ethereum address.
    :type address: str

    :returns: bool -- Ethereum valid/invalid address.

    >>> from swap.providers.ethereum.utils import is_address
    >>> is_address(address="0x1ee11011ae12103a488a82dc33e03f337bc93ba7")
    True
    """

    # Check parameter instances
    if not isinstance(address, str):
        raise TypeError(f"Address must be str, not '{type(address)}' type.")

    return Web3.isAddress(address)


def is_checksum_address(address: str) -> bool:
    """
    Check Ethereum checksum address.

    :param address: Ethereum address.
    :type address: str

    :returns: bool -- Ethereum valid/invalid checksum address.

    >>> from swap.providers.ethereum.utils import is_checksum_address
    >>> is_checksum_address(address="0x1ee11011ae12103a488a82dc33e03f337bc93ba7")
    False
    >>> is_checksum_address(address="0x1Ee11011ae12103a488A82DC33e03f337Bc93ba7")
    True
    """

    # Check parameter instances
    if not isinstance(address, str):
        raise TypeError(f"Address must be str, not '{type(address)}' type.")

    return Web3.isChecksumAddress(address)


def to_checksum_address(address: str) -> ChecksumAddress:
    """
    Change Ethereum address to checksum address.

    :param address: Ethereum address.
    :type address: ChecksumAddress

    :returns: str -- Ethereum checksum address.

    >>> from swap.providers.ethereum.utils import to_checksum_address
    >>>  is_checksum_address(address="0x1ee11011ae12103a488a82dc33e03f337bc93ba7")
    "0x1Ee11011ae12103a488A82DC33e03f337Bc93ba7"
    """

    # Check parameter instances
    if not is_address(address):
        raise AddressError(f"Invalid Ethereum '{type(address)}' address.")

    return Web3.toChecksumAddress(address)


def is_transaction_raw(transaction_raw: str) -> bool:
    """
    Check Ethereum transaction raw.

    :param transaction_raw: Ethereum transaction raw.
    :type transaction_raw: str

    :returns: bool -- Ethereum valid/invalid transaction raw.

    >>> from swap.providers.ethereum.utils import is_transaction_raw
    >>> transaction_raw: str = "eyJmZWUiOiAxMDAwMDAwMCwgImFkZHJlc3MiOiAiYm0xcWU5MHFqdDl3NG04cnQzdG51dTBwenAyNGRrZmZlbHlzOHpjd3llIiwgInJhdyI6ICIwNzAxMDAwMjAxNWYwMTVkMzA1YTI4ZDhkMzRiNDBjNjU5MzY4MTBmOWU5YzFmOGJjOWM3OTNlYzJlNzJjNzBmOTIwM2ZiYmViMGE1NmRiOWZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmY4MGFkZTIwNDAxMDExNjAwMTQwZTQzYTkyYTllOGFjYTc4OGViMTU1MWMzMTY0NDhjMmUzZjc4MjE1MDEwMDAxNWYwMTVkMjAyZmQyNTU3YjY3ZjFkZjhiOGFjZWYwNjZmNWQ0NmE4NTAwODE0MzliNDE5MzI1ZDU1ZGJkOTM0MWUxMWFjNGZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmY4MDg0YWY1ZjAxMDExNjAwMTRjOTVlMDkyY2FlYWVjZTM1YzU3M2U3MWUxMTA1NTU2ZDkyOWNmYzkwMjIwMTIwNTk5MDdmZGFkMGZmOTVmZWJhNDNhZWYzN2QyZTU1YzU3YjZlMTg2Y2QzYWQxN2M4M2U2YzgwYzY1ODIxOGI2NTAyMDEzYWZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmY5MDRlMDExNjAwMTRjOTVlMDkyY2FlYWVjZTM1YzU3M2U3MWUxMTA1NTU2ZDkyOWNmYzkwMDAwMTNjZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmUwOTQ5MDY0MDExNjAwMTRjOTVlMDkyY2FlYWVjZTM1YzU3M2U3MWUxMTA1NTU2ZDkyOWNmYzkwMDAiLCAiaGFzaCI6ICI3NzlmYzliOWNhNGRiMTVkNDFhYzgwNDNlZDRlNDFkYjg4NDU2ZjA1YzljZmJhMDQ5MzYyZWNlZmQ2MjY3ZmMzIiwgInVuc2lnbmVkX2RhdGFzIjogW3siZGF0YXMiOiBbIjMzZThkYThjZThlZjEzZmI0OTM4YTM3NGFlYTM2NjRlNGNkMmNkMDBmZGQ5ZDI5ODU5M2JkYmQ4NzJkNjZiODgiXSwgIm5ldHdvcmsiOiAibWFpbm5ldCIsICJwYXRoIjogbnVsbH0sIHsiZGF0YXMiOiBbIjc1ZTg3Yzc5MzNiNGRjNGE4N2UwNmZlZDMyM2U4NDI1ZTU0YTQ5NGZmODBkYzdmOGM0NTUyY2RiMGE2YmM3NGEiXSwgInB1YmxpY19rZXkiOiAiNTk5MDdmZGFkMGZmOTVmZWJhNDNhZWYzN2QyZTU1YzU3YjZlMTg2Y2QzYWQxN2M4M2U2YzgwYzY1ODIxOGI2NSIsICJuZXR3b3JrIjogIm1haW5uZXQiLCAicGF0aCI6ICJtLzQ0LzE1My8xLzAvMSJ9XSwgInNpZ25hdHVyZXMiOiBbXSwgIm5ldHdvcmsiOiAibWFpbm5ldCIsICJ0eXBlIjogImJ5dG9tX2NsYWltX3Vuc2lnbmVkIn0"
    >>> is_transaction_raw(transaction_raw=transaction_raw)
    True
    """

    # Check parameter instances
    if not isinstance(transaction_raw, str):
        raise TypeError(f"Transaction raw must be str, not '{type(transaction_raw)}' type.")

    try:
        transaction_raw = clean_transaction_raw(transaction_raw)
        decoded_transaction_raw = b64decode(transaction_raw.encode())
        loaded_transaction_raw = json.loads(decoded_transaction_raw.decode())
        return loaded_transaction_raw["type"] in [
            "ethereum_fund_unsigned", "ethereum_fund_signed",
            "ethereum_withdraw_unsigned", "ethereum_withdraw_signed",
            "ethereum_refund_unsigned", "ethereum_refund_signed"
        ]
    except:
        return False


def decode_transaction_raw(transaction_raw: str) -> dict:
    """
    Decode Ethereum transaction raw.

    :param transaction_raw: Ethereum transaction raw.
    :type transaction_raw: str

    :returns: dict -- Decoded ethereum transaction raw.

    >>> from swap.providers.ethereum.utils import decode_transaction_raw
    >>> transaction_raw: str = "eyJmZWUiOiAxMDAwMDAwMCwgImFkZHJlc3MiOiAiYm0xcWU5MHFqdDl3NG04cnQzdG51dTBwenAyNGRrZmZlbHlzOHpjd3llIiwgInJhdyI6ICIwNzAxMDAwMjAxNWYwMTVkMzA1YTI4ZDhkMzRiNDBjNjU5MzY4MTBmOWU5YzFmOGJjOWM3OTNlYzJlNzJjNzBmOTIwM2ZiYmViMGE1NmRiOWZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmY4MGFkZTIwNDAxMDExNjAwMTQwZTQzYTkyYTllOGFjYTc4OGViMTU1MWMzMTY0NDhjMmUzZjc4MjE1MDEwMDAxNWYwMTVkMjAyZmQyNTU3YjY3ZjFkZjhiOGFjZWYwNjZmNWQ0NmE4NTAwODE0MzliNDE5MzI1ZDU1ZGJkOTM0MWUxMWFjNGZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmY4MDg0YWY1ZjAxMDExNjAwMTRjOTVlMDkyY2FlYWVjZTM1YzU3M2U3MWUxMTA1NTU2ZDkyOWNmYzkwMjIwMTIwNTk5MDdmZGFkMGZmOTVmZWJhNDNhZWYzN2QyZTU1YzU3YjZlMTg2Y2QzYWQxN2M4M2U2YzgwYzY1ODIxOGI2NTAyMDEzYWZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmY5MDRlMDExNjAwMTRjOTVlMDkyY2FlYWVjZTM1YzU3M2U3MWUxMTA1NTU2ZDkyOWNmYzkwMDAwMTNjZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmUwOTQ5MDY0MDExNjAwMTRjOTVlMDkyY2FlYWVjZTM1YzU3M2U3MWUxMTA1NTU2ZDkyOWNmYzkwMDAiLCAiaGFzaCI6ICI3NzlmYzliOWNhNGRiMTVkNDFhYzgwNDNlZDRlNDFkYjg4NDU2ZjA1YzljZmJhMDQ5MzYyZWNlZmQ2MjY3ZmMzIiwgInVuc2lnbmVkX2RhdGFzIjogW3siZGF0YXMiOiBbIjMzZThkYThjZThlZjEzZmI0OTM4YTM3NGFlYTM2NjRlNGNkMmNkMDBmZGQ5ZDI5ODU5M2JkYmQ4NzJkNjZiODgiXSwgIm5ldHdvcmsiOiAibWFpbm5ldCIsICJwYXRoIjogbnVsbH0sIHsiZGF0YXMiOiBbIjc1ZTg3Yzc5MzNiNGRjNGE4N2UwNmZlZDMyM2U4NDI1ZTU0YTQ5NGZmODBkYzdmOGM0NTUyY2RiMGE2YmM3NGEiXSwgInB1YmxpY19rZXkiOiAiNTk5MDdmZGFkMGZmOTVmZWJhNDNhZWYzN2QyZTU1YzU3YjZlMTg2Y2QzYWQxN2M4M2U2YzgwYzY1ODIxOGI2NSIsICJuZXR3b3JrIjogIm1haW5uZXQiLCAicGF0aCI6ICJtLzQ0LzE1My8xLzAvMSJ9XSwgInNpZ25hdHVyZXMiOiBbXSwgIm5ldHdvcmsiOiAibWFpbm5ldCIsICJ0eXBlIjogImJ5dG9tX2NsYWltX3Vuc2lnbmVkIn0"
    >>> decode_transaction_raw(transaction_raw=transaction_raw)
    {'fee': ..., 'type': '...', 'address': '...', 'transaction': {...}, 'unsigned_datas': [...], 'signatures': [...], 'network': '...'}
    """

    if not is_transaction_raw(transaction_raw=transaction_raw):
        raise TransactionRawError("Invalid Ethereum transaction raw.")

    transaction_raw = clean_transaction_raw(transaction_raw)
    decoded_transaction_raw = b64decode(transaction_raw.encode())
    loaded_transaction_raw = json.loads(decoded_transaction_raw.decode())

    return dict(
        fee=loaded_transaction_raw["fee"],
        transaction=loaded_transaction_raw["transaction"],
        signatures=loaded_transaction_raw["signature"],
        network=loaded_transaction_raw["network"],
        type=loaded_transaction_raw["type"],
    )


def submit_transaction_raw(transaction_raw: str, provider: str = config["provider"],
                           token: Optional[str] = None) -> dict:
    """
    Submit Ethereum transaction raw.

    :param transaction_raw: Ethereum transaction raw.
    :type transaction_raw: str
    :param provider: Ethereum network provider, defaults to ``http``.
    :type provider: str
    :param token: Infura API endpoint token, defaults to ``4414fea5f7454211956b1627621450b4``.
    :type token: str

    :returns: dict -- Ethereum submitted transaction id, fee, type and date.

    >>> from swap.providers.ethereum.utils import submit_transaction_raw
    >>> transaction_raw: str = "eyJmZWUiOiAxMDAwMDAwMCwgImFkZHJlc3MiOiAiYm0xcWU5MHFqdDl3NG04cnQzdG51dTBwenAyNGRrZmZlbHlzOHpjd3llIiwgInJhdyI6ICIwNzAxMDAwMjAxNWYwMTVkMzA1YTI4ZDhkMzRiNDBjNjU5MzY4MTBmOWU5YzFmOGJjOWM3OTNlYzJlNzJjNzBmOTIwM2ZiYmViMGE1NmRiOWZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmY4MGFkZTIwNDAxMDExNjAwMTQwZTQzYTkyYTllOGFjYTc4OGViMTU1MWMzMTY0NDhjMmUzZjc4MjE1MDEwMDAxNWYwMTVkMjAyZmQyNTU3YjY3ZjFkZjhiOGFjZWYwNjZmNWQ0NmE4NTAwODE0MzliNDE5MzI1ZDU1ZGJkOTM0MWUxMWFjNGZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmY4MDg0YWY1ZjAxMDExNjAwMTRjOTVlMDkyY2FlYWVjZTM1YzU3M2U3MWUxMTA1NTU2ZDkyOWNmYzkwMjIwMTIwNTk5MDdmZGFkMGZmOTVmZWJhNDNhZWYzN2QyZTU1YzU3YjZlMTg2Y2QzYWQxN2M4M2U2YzgwYzY1ODIxOGI2NTAyMDEzYWZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmY5MDRlMDExNjAwMTRjOTVlMDkyY2FlYWVjZTM1YzU3M2U3MWUxMTA1NTU2ZDkyOWNmYzkwMDAwMTNjZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmZmUwOTQ5MDY0MDExNjAwMTRjOTVlMDkyY2FlYWVjZTM1YzU3M2U3MWUxMTA1NTU2ZDkyOWNmYzkwMDAiLCAiaGFzaCI6ICI3NzlmYzliOWNhNGRiMTVkNDFhYzgwNDNlZDRlNDFkYjg4NDU2ZjA1YzljZmJhMDQ5MzYyZWNlZmQ2MjY3ZmMzIiwgInVuc2lnbmVkX2RhdGFzIjogW3siZGF0YXMiOiBbIjMzZThkYThjZThlZjEzZmI0OTM4YTM3NGFlYTM2NjRlNGNkMmNkMDBmZGQ5ZDI5ODU5M2JkYmQ4NzJkNjZiODgiXSwgIm5ldHdvcmsiOiAibWFpbm5ldCIsICJwYXRoIjogbnVsbH0sIHsiZGF0YXMiOiBbIjc1ZTg3Yzc5MzNiNGRjNGE4N2UwNmZlZDMyM2U4NDI1ZTU0YTQ5NGZmODBkYzdmOGM0NTUyY2RiMGE2YmM3NGEiXSwgInB1YmxpY19rZXkiOiAiNTk5MDdmZGFkMGZmOTVmZWJhNDNhZWYzN2QyZTU1YzU3YjZlMTg2Y2QzYWQxN2M4M2U2YzgwYzY1ODIxOGI2NSIsICJuZXR3b3JrIjogIm1haW5uZXQiLCAicGF0aCI6ICJtLzQ0LzE1My8xLzAvMSJ9XSwgInNpZ25hdHVyZXMiOiBbXSwgIm5ldHdvcmsiOiAibWFpbm5ldCIsICJ0eXBlIjogImJ5dG9tX2NsYWltX3Vuc2lnbmVkIn0"
    >>> submit_transaction_raw(transaction_raw=transaction_raw)
    {'fee': ..., 'type': '...', 'transaction_id': '...', 'network': '...', 'date': '...'}
    """

    # Check parameter instances
    if not is_transaction_raw(transaction_raw=transaction_raw):
        raise TransactionRawError("Invalid Ethereum transaction raw.")

    transaction_raw = clean_transaction_raw(transaction_raw)
    decoded_transaction_raw = b64decode(transaction_raw.encode())
    loaded_transaction_raw = json.loads(decoded_transaction_raw.decode())

    if not loaded_transaction_raw["type"] in [
        "ethereum_fund_signed", "ethereum_withdraw_signed", "ethereum_refund_signed"
    ]:
        raise TransactionRawError("Wrong Ethereum transaction raw must be signed, not unsigned transaction raw.")

    from .rpc import get_web3
    web3: Web3 = get_web3(
        network=loaded_transaction_raw["network"], provider=provider, token=token
    )
    transaction_hash: HexBytes = web3.eth.send_raw_transaction(
        loaded_transaction_raw["signature"]["rawTransaction"]
    )

    return dict(
        fee=loaded_transaction_raw["fee"],
        type=loaded_transaction_raw["type"],
        transaction_hash=transaction_hash.hex(),
        network=loaded_transaction_raw["network"],
        date=str(datetime.now())
    )


def amount_unit_converter(amount: Union[int, float], unit_from: str = "Wei2Ether") -> Union[int, float]:
    """
    XinFin amount unit converter.

    :param amount: XinFIn amount.
    :type amount: int, float
    :param unit_from: XinFIn unit, default to Wei2Ether
    :type unit_from: str

    :returns: int, float -- XinFin amount.

    >>> from swap.providers.ethereum.utils import amount_unit_converter
    >>> amount_unit_converter(amount=100_000_000, unit_from="Wei2Ether")
    0.1
    """

    if unit_from not in ["Ether2Gwei", "Ether2Wei", "Gwei2Ether", "Gwei2Wei", "Wei2Ether", "Wei2Gwei"]:
        raise UnitError(f"Invalid Ethereum '{unit_from}' unit from",
                        "choose only 'Ether2Gwei', 'Ether2Wei', 'Gwei2Ether', 'Gwei2Wei', 'Wei2Ether' or 'Wei2Gwei' units.")

    # Constant values
    Ether, Gwei, Wei = (
        config["units"]["Ether"],
        config["units"]["Gwei"],
        config["units"]["Wei"]
    )

    if unit_from == "Ether2Gwei":
        return float((amount * Gwei) / Ether)
    elif unit_from == "Ether2Wei":
        return int((amount * Wei) / Ether)
    elif unit_from == "Gwei2Ether":
        return float((amount * Ether) / Gwei)
    elif unit_from == "Gwei2Wei":
        return int((amount * Wei) / Gwei)
    elif unit_from == "Wei2Ether":
        return float((amount * Ether) / Wei)
    elif unit_from == "Wei2Gwei":
        return int((amount * Gwei) / Wei)



#!/usr/bin/env python3

import json
import os

from swap.cli.__main__ import main as cli_main
from swap.utils import clean_transaction_raw

# Test Values
base_path = os.path.dirname(__file__)
file_path = os.path.abspath(os.path.join(base_path, "..", "..", "values.json"))
values = open(file_path, "r")
_ = json.loads(values.read())
values.close()


def test_bytom_cli_withdraw(cli_tester):

    withdraw = cli_tester.invoke(
        cli_main, [
            "bytom",
            "withdraw",
            "--address", _["bytom"]["wallet"]["recipient"]["address"],
            "--transaction-hash", _["bytom"]["transaction_hash"],
            "--asset", _["bytom"]["asset"],
            "--network", _["bytom"]["network"]
        ]
    )
    assert withdraw.exit_code == 0
    assert withdraw.output == clean_transaction_raw(
        transaction_raw=_["bytom"]["withdraw"]["unsigned"]["transaction_raw"]
    ) + "\n"



#!/usr/bin/env python3

import json
import os

from swap.providers.xinfin.wallet import Wallet

# Test Values
base_path = os.path.dirname(__file__)
file_path = os.path.abspath(os.path.join(base_path, "..", "..", "values.json"))
values = open(file_path, "r")
_ = json.loads(values.read())
values.close()


def test_xinfin_wallet_from_entropy():

    wallet = Wallet(network=_["xinfin"]["network"])

    wallet.from_entropy(
        entropy=_["xinfin"]["wallet"]["sender"]["entropy"],
        language=_["xinfin"]["wallet"]["sender"]["language"],
        passphrase=_["xinfin"]["wallet"]["sender"]["passphrase"]
    )

    wallet.from_path(
        path=_["xinfin"]["wallet"]["sender"]["derivation"]["path"]
    )

    assert wallet.entropy() == _["xinfin"]["wallet"]["sender"]["entropy"]
    assert wallet.mnemonic() == _["xinfin"]["wallet"]["sender"]["mnemonic"]
    assert wallet.language() == _["xinfin"]["wallet"]["sender"]["language"]
    assert wallet.passphrase() is None
    assert wallet.seed() == _["xinfin"]["wallet"]["sender"]["seed"]
    assert wallet.root_xprivate_key() == _["xinfin"]["wallet"]["sender"]["root_xprivate_key"]
    assert wallet.root_xpublic_key() == _["xinfin"]["wallet"]["sender"]["root_xpublic_key"]
    assert wallet.xprivate_key() == _["xinfin"]["wallet"]["sender"]["xprivate_key"]
    assert wallet.xpublic_key() == _["xinfin"]["wallet"]["sender"]["xpublic_key"]
    assert wallet.uncompressed() == _["xinfin"]["wallet"]["sender"]["uncompressed"]
    assert wallet.compressed() == _["xinfin"]["wallet"]["sender"]["compressed"]
    assert wallet.chain_code() == _["xinfin"]["wallet"]["sender"]["chain_code"]
    assert wallet.private_key() == _["xinfin"]["wallet"]["sender"]["private_key"]
    assert wallet.public_key() == _["xinfin"]["wallet"]["sender"]["public_key"]
    assert wallet.wif() == _["xinfin"]["wallet"]["sender"]["wif"]
    assert wallet.hash() == _["xinfin"]["wallet"]["sender"]["hash"]
    assert wallet.finger_print() == _["xinfin"]["wallet"]["sender"]["finger_print"]
    assert wallet.path() == _["xinfin"]["wallet"]["sender"]["derivation"]["path"]
    assert wallet.address() == _["xinfin"]["wallet"]["sender"]["address"]

    assert isinstance(wallet.balance(), int)


def test_xinfin_wallet_from_mnemonic():

    wallet = Wallet(network=_["xinfin"]["network"])

    wallet.from_mnemonic(
        mnemonic=_["xinfin"]["wallet"]["sender"]["mnemonic"],
        language=_["xinfin"]["wallet"]["sender"]["language"],
        passphrase=_["xinfin"]["wallet"]["sender"]["passphrase"]
    )

    wallet.from_path(
        path=_["xinfin"]["wallet"]["sender"]["derivation"]["path"]
    )

    assert wallet.entropy() == _["xinfin"]["wallet"]["sender"]["entropy"]
    assert wallet.mnemonic() == _["xinfin"]["wallet"]["sender"]["mnemonic"]
    assert wallet.language() == _["xinfin"]["wallet"]["sender"]["language"]
    assert wallet.passphrase() is None
    assert wallet.seed() == _["xinfin"]["wallet"]["sender"]["seed"]
    assert wallet.root_xprivate_key() == _["xinfin"]["wallet"]["sender"]["root_xprivate_key"]
    assert wallet.root_xpublic_key() == _["xinfin"]["wallet"]["sender"]["root_xpublic_key"]
    assert wallet.xprivate_key() == _["xinfin"]["wallet"]["sender"]["xprivate_key"]
    assert wallet.xpublic_key() == _["xinfin"]["wallet"]["sender"]["xpublic_key"]
    assert wallet.uncompressed() == _["xinfin"]["wallet"]["sender"]["uncompressed"]
    assert wallet.compressed() == _["xinfin"]["wallet"]["sender"]["compressed"]
    assert wallet.chain_code() == _["xinfin"]["wallet"]["sender"]["chain_code"]
    assert wallet.private_key() == _["xinfin"]["wallet"]["sender"]["private_key"]
    assert wallet.public_key() == _["xinfin"]["wallet"]["sender"]["public_key"]
    assert wallet.wif() == _["xinfin"]["wallet"]["sender"]["wif"]
    assert wallet.hash() == _["xinfin"]["wallet"]["sender"]["hash"]
    assert wallet.finger_print() == _["xinfin"]["wallet"]["sender"]["finger_print"]
    assert wallet.path() == _["xinfin"]["wallet"]["sender"]["derivation"]["path"]
    assert wallet.address() == _["xinfin"]["wallet"]["sender"]["address"]

    # assert isinstance(wallet.balance(), int)


def test_xinfin_wallet_from_seed():

    wallet = Wallet(network=_["xinfin"]["network"])

    wallet.from_seed(
        seed=_["xinfin"]["wallet"]["recipient"]["seed"]
    )

    wallet.from_path(
        path=_["xinfin"]["wallet"]["recipient"]["derivation"]["path"]
    )

    assert wallet.entropy() is None
    assert wallet.mnemonic() is None
    assert wallet.language() is None
    assert wallet.passphrase() is None
    assert wallet.seed() == _["xinfin"]["wallet"]["recipient"]["seed"]
    assert wallet.root_xprivate_key() == _["xinfin"]["wallet"]["recipient"]["root_xprivate_key"]
    assert wallet.root_xpublic_key() == _["xinfin"]["wallet"]["recipient"]["root_xpublic_key"]
    assert wallet.xprivate_key() == _["xinfin"]["wallet"]["recipient"]["xprivate_key"]
    assert wallet.xpublic_key() == _["xinfin"]["wallet"]["recipient"]["xpublic_key"]
    assert wallet.uncompressed() == _["xinfin"]["wallet"]["recipient"]["uncompressed"]
    assert wallet.compressed() == _["xinfin"]["wallet"]["recipient"]["compressed"]
    assert wallet.chain_code() == _["xinfin"]["wallet"]["recipient"]["chain_code"]
    assert wallet.private_key() == _["xinfin"]["wallet"]["recipient"]["private_key"]
    assert wallet.public_key() == _["xinfin"]["wallet"]["recipient"]["public_key"]
    assert wallet.wif() == _["xinfin"]["wallet"]["recipient"]["wif"]
    assert wallet.hash() == _["xinfin"]["wallet"]["recipient"]["hash"]
    assert wallet.finger_print() == _["xinfin"]["wallet"]["recipient"]["finger_print"]
    assert wallet.path() == _["xinfin"]["wallet"]["recipient"]["derivation"]["path"]
    assert wallet.address() == _["xinfin"]["wallet"]["recipient"]["address"]

    # assert isinstance(wallet.balance(), int)


def test_xinfin_wallet_from_root_xprivate_key():

    wallet = Wallet(network=_["xinfin"]["network"])

    wallet.from_root_xprivate_key(
        xprivate_key=_["xinfin"]["wallet"]["sender"]["root_xprivate_key"]
    )

    wallet.from_path(
        path=_["xinfin"]["wallet"]["sender"]["derivation"]["path"]
    )

    assert wallet.entropy() is None
    assert wallet.mnemonic() is None
    assert wallet.language() is None
    assert wallet.passphrase() is None
    assert wallet.seed() is None
    assert wallet.root_xprivate_key() == _["xinfin"]["wallet"]["sender"]["root_xprivate_key"]
    assert wallet.root_xpublic_key() == _["xinfin"]["wallet"]["sender"]["root_xpublic_key"]
    assert wallet.xprivate_key() == _["xinfin"]["wallet"]["sender"]["xprivate_key"]
    assert wallet.xpublic_key() == _["xinfin"]["wallet"]["sender"]["xpublic_key"]
    assert wallet.uncompressed() == _["xinfin"]["wallet"]["sender"]["uncompressed"]
    assert wallet.compressed() == _["xinfin"]["wallet"]["sender"]["compressed"]
    assert wallet.chain_code() == _["xinfin"]["wallet"]["sender"]["chain_code"]
    assert wallet.private_key() == _["xinfin"]["wallet"]["sender"]["private_key"]
    assert wallet.public_key() == _["xinfin"]["wallet"]["sender"]["public_key"]
    assert wallet.wif() == _["xinfin"]["wallet"]["sender"]["wif"]
    assert wallet.hash() == _["xinfin"]["wallet"]["sender"]["hash"]
    assert wallet.finger_print() == _["xinfin"]["wallet"]["sender"]["finger_print"]
    assert wallet.path() == _["xinfin"]["wallet"]["sender"]["derivation"]["path"]
    assert wallet.address() == _["xinfin"]["wallet"]["sender"]["address"]

    # assert isinstance(wallet.balance(), int)


def test_xinfin_wallet_from_xprivate_key():

    wallet = Wallet(network=_["xinfin"]["network"])

    wallet.from_xprivate_key(
        xprivate_key=_["xinfin"]["wallet"]["recipient"]["xprivate_key"]
    )

    assert wallet.entropy() is None
    assert wallet.mnemonic() is None
    assert wallet.language() is None
    assert wallet.passphrase() is None
    assert wallet.seed() is None
    assert wallet.root_xprivate_key() is None
    assert wallet.root_xpublic_key() is None
    assert wallet.xprivate_key() == _["xinfin"]["wallet"]["recipient"]["xprivate_key"]
    assert wallet.xpublic_key() == _["xinfin"]["wallet"]["recipient"]["xpublic_key"]
    assert wallet.uncompressed() == _["xinfin"]["wallet"]["recipient"]["uncompressed"]
    assert wallet.compressed() == _["xinfin"]["wallet"]["recipient"]["compressed"]
    assert wallet.chain_code() == _["xinfin"]["wallet"]["recipient"]["chain_code"]
    assert wallet.private_key() == _["xinfin"]["wallet"]["recipient"]["private_key"]
    assert wallet.public_key() == _["xinfin"]["wallet"]["recipient"]["public_key"]
    assert wallet.wif() == _["xinfin"]["wallet"]["recipient"]["wif"]
    assert wallet.hash() == _["xinfin"]["wallet"]["recipient"]["hash"]
    assert wallet.finger_print() == _["xinfin"]["wallet"]["recipient"]["finger_print"]
    assert wallet.path() is None
    assert wallet.address() == _["xinfin"]["wallet"]["recipient"]["address"]

    # assert isinstance(wallet.balance(), int)


def test_xinfin_wallet_from_private_key():

    wallet = Wallet(network=_["xinfin"]["network"])

    wallet.from_private_key(
        private_key=_["xinfin"]["wallet"]["sender"]["private_key"]
    )

    assert wallet.entropy() is None
    assert wallet.mnemonic() is None
    assert wallet.language() is None
    assert wallet.passphrase() is None
    assert wallet.seed() is None
    assert wallet.root_xprivate_key() is None
    assert wallet.root_xpublic_key() is None
    assert wallet.xprivate_key() is None
    assert wallet.xpublic_key() is None
    assert wallet.uncompressed() == _["xinfin"]["wallet"]["sender"]["uncompressed"]
    assert wallet.compressed() == _["xinfin"]["wallet"]["sender"]["compressed"]
    assert wallet.chain_code() is None
    assert wallet.private_key() == _["xinfin"]["wallet"]["sender"]["private_key"]
    assert wallet.public_key() == _["xinfin"]["wallet"]["sender"]["public_key"]
    assert wallet.wif() == _["xinfin"]["wallet"]["sender"]["wif"]
    assert wallet.hash() == _["xinfin"]["wallet"]["sender"]["hash"]
    assert wallet.finger_print() == _["xinfin"]["wallet"]["sender"]["finger_print"]
    assert wallet.path() is None
    assert wallet.address() == _["xinfin"]["wallet"]["sender"]["address"]

    # assert isinstance(wallet.balance(), int)



#!/usr/bin/env python3

from hdwallet import HDWallet
from hdwallet.cryptocurrencies import (
    BitcoinMainnet, BitcoinTestnet
)
from typing import (
    Optional, Any, Union
)

from ...utils import is_mnemonic
from ...exceptions import (
    NetworkError, UnitError
)
from ..config import bitcoin as config
from .utils import (
    is_network, amount_unit_converter
)
from .rpc import (
    get_balance, get_utxos
)

# Default derivation path
DEFAULT_PATH: str = config["path"]


class Wallet(HDWallet):
    """
    Bitcoin hierarchical deterministic wallet.

    :param network: Bitcoin network, defaults to ``mainnet``.
    :type network: str
    :param use_default_path: Use default derivation path, defaults to ``False``.
    :type use_default_path: bool

    :returns: Wallet -- Bitcoin instance.

    .. note::
        Bitcoin has only two networks, ``mainnet`` and ``mainnet``.
    """

    def __init__(self, network: str = config["network"], use_default_path: bool = False):

        # Check parameter instances
        if not is_network(network=network):
            raise NetworkError(f"Invalid Bitcoin '{network}' network",
                               "choose only 'mainnet' or 'testnet' networks.")

        self._network: str = network
        self._cryptocurrency: Any = (
            BitcoinMainnet if self._network == "mainnet" else BitcoinTestnet
        )
        self._hdwallet: HDWallet = HDWallet(
            cryptocurrency=self._cryptocurrency, use_default_path=use_default_path
        )

        super().__init__(cryptocurrency=self._cryptocurrency)

    def from_entropy(self, entropy: str, language: str = "english", passphrase: Optional[str] = None) -> "Wallet":
        """
        Initialize wallet from entropy.

        :param entropy: Bitcoin entropy.
        :type entropy: str
        :param language: Bitcoin language, default to english.
        :type language: str
        :param passphrase: Bitcoin passphrase, default to None.
        :type passphrase: str

        :returns: Wallet -- Bitcoin instance.

        >>> from swap.providers.bitcoin.wallet import Wallet
        >>> wallet: Wallet = Wallet(network="testnet")
        >>> wallet.from_entropy(entropy="ed0802d701a033776811601dd6c5c4a9")
        <swap.providers.bitcoin.wallet.Wallet object at 0x040DA268>
        """

        self._hdwallet.from_entropy(entropy=entropy, language=language, passphrase=passphrase)
        return self

    def from_mnemonic(self, mnemonic: str, language: Optional[str] = None,
                      passphrase: Optional[str] = None) -> "Wallet":
        """
        Initialize wallet from mnemonic.

        :param mnemonic: Bitcoin mnemonic.
        :type mnemonic: str
        :param language: Bitcoin language, default to english.
        :type language: str
        :param passphrase: Bitcoin passphrase, default to None.
        :type passphrase: str

        :returns: Wallet -- Bitcoin instance.

        >>> from swap.providers.bitcoin.wallet import Wallet
        >>> wallet: Wallet = Wallet(network="testnet")
        >>> wallet.from_mnemonic(mnemonic="unfair divorce remind addict add roof park clown build renew illness fault")
        <swap.providers.bitcoin.wallet.Wallet object at 0x040DA268>
        """

        # Check parameter instances
        if not is_mnemonic(mnemonic=mnemonic, language=language):
            raise ValueError("Invalid Mnemonic words.")

        self._hdwallet.from_mnemonic(mnemonic=mnemonic, language=language, passphrase=passphrase)
        return self

    def from_seed(self, seed: str) -> "Wallet":
        """
        Initialize wallet from seed.

        :param seed: Bitcoin seed.
        :type seed: str

        :returns: Wallet -- Bitcoin instance.

        >>> from swap.providers.bitcoin.wallet import Wallet
        >>> wallet: Wallet = Wallet(network="testnet")
        >>> wallet.from_seed(seed="1cfd5df8a523d53a36cee369a93fac4e9efab5e4e138d479da2fb6df730697574409d572fe8325ec22e8ed25dea7495f498c3f5235fe6ae6d47b989267b6777c")
        <swap.providers.bitcoin.wallet.Wallet object at 0x040DA268>
        """

        self._hdwallet.from_seed(seed=seed)
        return self

    def from_root_xprivate_key(self, xprivate_key: str, strict: bool = True) -> "Wallet":
        """
        Master from Root XPrivate Key.

        :param xprivate_key: Bitcoin XPrivate key.
        :type xprivate_key: str
        :param strict: Strict for must be root xprivate key, default to ``True``.
        :type strict: bool

        :returns: Wallet -- Bitcoin instance.

        >>> from swap.providers.bitcoin.wallet import Wallet
        >>> wallet: Wallet = Wallet(network="testnet")
        >>> wallet.from_xprivate_key(xprivate_key="tprv8ZgxMBicQKsPeMHMJAc6uWGYiGqi1MVM2ybmzXL2TAoDpQe85uyDpdT7mv7Nhdu5rTCBEKLZsd9KyP2LQZJzZTvgVQvENArgU8e6DoYBiXf")
        <swap.providers.bitcoin.wallet.Wallet object at 0x040DA268>
        """

        self._hdwallet.from_root_xprivate_key(xprivate_key=xprivate_key, strict=strict)
        return self

    def from_root_xpublic_key(self, xpublic_key: str, strict: bool = True) -> "Wallet":
        """
        Master from Root XPublic Key.

        :param xpublic_key: Bitcoin XPublic key.
        :type xpublic_key: str
        :param strict: Strict for must be root xprivate key, default to ``True``.
        :type strict: bool

        :returns: Wallet -- Bitcoin instance.

        >>> from swap.providers.bitcoin.wallet import Wallet
        >>> wallet: Wallet = Wallet(network="testnet")
        >>> wallet.from_root_xpublic_key(xpublic_key="tpubD6NzVbkrYhZ4XpK9BpGhJuvfHJMeAggFcHCZH3NKsSbcetttiJnp184yx2cp2uJyapPQLt7LGTLUZvnKWbdgKBkvnfYjab9sH4wBmEpTZhJ")
        <swap.providers.bitcoin.wallet.Wallet object at 0x040DA268>
        """

        self._hdwallet.from_root_xpublic_key(xpublic_key=xpublic_key, strict=strict)
        return self

    def from_xprivate_key(self, xprivate_key: str) -> "Wallet":
        """
        Initialize wallet from root xprivate key.

        :param xprivate_key: Bitcoin root xprivate key.
        :type xprivate_key: str

        :returns: Wallet -- Bitcoin instance.

        >>> from swap.providers.bitcoin.wallet import Wallet
        >>> wallet: Wallet = Wallet(network="testnet")
        >>> wallet.from_xprivate_key(xprivate_key="tprv8kqWVfMdSgo9WhUAxbmL6GNW4ivePvEZBu8QiiRfMXbVDgnHx16vndnAsv7Uds4iFvjMpdJiB6q6hhh753fRb89XFjHGjYJ8BsMZGv3RTKz")
        <swap.providers.bitcoin.wallet.Wallet object at 0x040DA268>
        """

        self._hdwallet.from_xprivate_key(xprivate_key=xprivate_key)
        return self

    def from_xpublic_key(self, xpublic_key: str) -> "Wallet":
        """
        Initialize wallet from XPrivate key.

        :param xpublic_key: Bitcoin XPrivate key.
        :type xpublic_key: str

        :returns: Wallet -- Bitcoin instance.

        >>> from swap.providers.bitcoin.wallet import Wallet
        >>> wallet: Wallet = Wallet(network="testnet")
        >>> wallet.from_xprivate_key(xprivate_key="tpubDHXYe5Psb4UpQAVxrFRvVg2cdkSaZFRTmCjC1ETxmoPt4B34aPvWy8Q343tUsTaCQCiSJVpzgyP1NQ3mffY7oF6u6cJ6Csx3AhgFFLVoUBu")
        <swap.providers.bitcoin.wallet.Wallet object at 0x040DA268>
        """

        self._hdwallet.from_xpublic_key(xpublic_key=xpublic_key)
        return self

    def from_wif(self, wif: str) -> "Wallet":
        """
        Initialize wallet from wallet important format (WIF).

        :param wif: Bitcoin important format.
        :type wif: str

        :returns: Wallet -- Bitcoin instance.

        >>> from swap.providers.bitcoin.wallet import Wallet
        >>> wallet: Wallet = Wallet(network="testnet")
        >>> wallet.from_wif(wif="cS6utJFQYTQEAY455hRQ5nardhCCoc2yf4M45P71ve5Dx44ag7qg")
        <swap.providers.bitcoin.wallet.Wallet object at 0x040DA268>
        """

        self._hdwallet.from_wif(wif=wif)
        return self

    def from_private_key(self, private_key) -> "Wallet":
        """
        Initialize wallet from private key.

        :param private_key: Bitcoin private key.
        :type private_key: str

        :returns: Wallet -- Bitcoin instance.

        >>> from swap.providers.bitcoin.wallet import Wallet
        >>> wallet: Wallet = Wallet(network="testnet")
        >>> wallet.from_private_key(private_key="86e4296c4b8804b952933ddf9b786a0bad1049c1d5b372e43f9336eb4ac2fcb6")
        <swap.providers.bitcoin.wallet.Wallet object at 0x040DA268>
        """

        self._hdwallet.from_private_key(private_key=private_key)
        return self

    def from_path(self, path: str) -> "Wallet":
        """
        Drive Bitcoin from path.

        :param path: Bitcoin path.
        :type path: str

        :returns: Wallet -- Bitcoin instance.

        >>> from swap.providers.bitcoin.wallet import Wallet
        >>> wallet: Wallet = Wallet(network="testnet", use_default_path=False)
        >>> wallet.from_entropy("ed0802d701a033776811601dd6c5c4a9")
        >>> wallet.from_path(path="m/44'/1'/0'/0/0")
        <swap.providers.bitcoin.wallet.Wallet object at 0x040DA268>
        """

        self._hdwallet.from_path(path=path)
        return self

    def from_index(self, index: int, hardened: bool = False) -> "Wallet":
        """
        Drive Bitcoin from index.

        :param index: Bitcoin index.
        :type index: int
        :param hardened: Use harden, default to False.
        :type hardened: bool

        :returns: Wallet -- Bitcoin instance.

        >>> from swap.providers.bitcoin.wallet import Wallet
        >>> wallet: Wallet = Wallet(network="testnet", use_default_path=False)
        >>> wallet.from_entropy("ed0802d701a033776811601dd6c5c4a9")
        >>> wallet.from_index(index=44, hardened=True)
        >>> wallet.from_index(index=1, hardened=True)
        >>> wallet.from_index(index=0, hardened=True)
        >>> wallet.from_index(index=0)
        >>> wallet.from_index(index=0)
        <swap.providers.bitcoin.wallet.Wallet object at 0x040DA268>
        """

        self._hdwallet.from_index(index=index, hardened=hardened)
        return self

    def clean_derivation(self) -> "Wallet":
        """
        Clean derivation Bitcoin.

        :returns: Wallet -- Bitcoin instance.

        >>> from swap.providers.bitcoin.wallet import Wallet
        >>> wallet: Wallet = Wallet(network="testnet")
        >>> wallet.from_entropy("ed0802d701a033776811601dd6c5c4a9")
        >>> wallet.from_path("m/44'/1'/0'/0/0")
        >>> wallet.path()
        "m/44'/1'/0'/0/0"
        >>> wallet.clean_derivation()
        <swap.providers.bitcoin.wallet.Wallet object at 0x040DA268>
        >>> wallet.path()
        None
        """

        self._hdwallet.clean_derivation()
        return self

    def strength(self) -> Optional[int]:
        """
        Get Bitcoin strength.

        :return: int -- Bitcoin strength.

        >>> from swap.providers.bitcoin.wallet import Wallet
        >>> wallet: Wallet = Wallet(network="testnet")
        >>> wallet.from_entropy("ed0802d701a033776811601dd6c5c4a9")
        >>> wallet.strength()
        128
        """

        return self._hdwallet.strength()

    def entropy(self) -> Optional[str]:
        """
        Get Bitcoin entropy.

        :return: str -- Bitcoin entropy.

        >>> from swap.providers.bitcoin.wallet import Wallet
        >>> wallet: Wallet = Wallet(network="testnet")
        >>> wallet.from_mnemonic(mnemonic="unfair divorce remind addict add roof park clown build renew illness fault")
        >>> wallet.entropy()
        "ed0802d701a033776811601dd6c5c4a9"
        """

        return self._hdwallet.entropy()

    def mnemonic(self) -> Optional[str]:
        """
        Get Bitcoin mnemonic.

        :return: str -- Bitcoin mnemonic.

        >>> from swap.providers.bitcoin.wallet import Wallet
        >>> wallet: Wallet = Wallet(network="testnet")
        >>> wallet.from_entropy("ed0802d701a033776811601dd6c5c4a9")
        >>> wallet.mnemonic()
        "unfair divorce remind addict add roof park clown build renew illness fault"
        """

        return self._hdwallet.mnemonic()

    def passphrase(self) -> Optional[str]:
        """
        Get Bitcoin passphrase.

        :return: str -- Bitcoin passphrase.

        >>> from swap.providers.bitcoin.wallet import Wallet
        >>> wallet: Wallet = Wallet(network="testnet")
        >>> wallet.from_entropy("ed0802d701a033776811601dd6c5c4a9", passphrase="meherett")
        >>> wallet.passphrase()
        "meherett"
        """

        return self._hdwallet.passphrase()

    def language(self) -> Optional[str]:
        """
        Get Bitcoin language.

        :return: str -- Bitcoin language.

        >>> from swap.providers.bitcoin.wallet import Wallet
        >>> wallet: Wallet = Wallet(network="testnet")
        >>> wallet.from_entropy("ed0802d701a033776811601dd6c5c4a9")
        >>> wallet.language()
        "english"
        """

        return self._hdwallet.language()

    def seed(self) -> Optional[str]:
        """
        Get Bitcoin seed.

        :return: str -- Bitcoin seed.

        >>> from swap.providers.bitcoin.wallet import Wallet
        >>> wallet: Wallet = Wallet(network="testnet")
        >>> wallet.from_entropy("ed0802d701a033776811601dd6c5c4a9")
        >>> wallet.seed()
        "1cfd5df8a523d53a36cee369a93fac4e9efab5e4e138d479da2fb6df730697574409d572fe8325ec22e8ed25dea7495f498c3f5235fe6ae6d47b989267b6777c"
        """

        return self._hdwallet.seed()

    def root_xprivate_key(self, encoded: bool = True) -> Optional[str]:
        """
        Get Bitcoin root xprivate key.

        :param encoded: Encoded root xprivate key, default to True.
        :type encoded: bool

        :return: str -- Bitcoin root xprivate key.

        >>> from swap.providers.bitcoin.wallet import Wallet
        >>> wallet: Wallet = Wallet(network="testnet")
        >>> wallet.from_entropy("ed0802d701a033776811601dd6c5c4a9")
        >>> wallet.root_xprivate_key()
        "tprv8ZgxMBicQKsPeMHMJAc6uWGYiGqi1MVM2ybmzXL2TAoDpQe85uyDpdT7mv7Nhdu5rTCBEKLZsd9KyP2LQZJzZTvgVQvENArgU8e6DoYBiXf"
        """

        return self._hdwallet.root_xprivate_key(encoded=encoded)

    def root_xpublic_key(self, encoded: bool = True) -> Optional[str]:
        """
        Get Bitcoin root xpublic key.

        :param encoded: Encoded root xpublic key, default to True.
        :type encoded: bool

        :return: str -- Bitcoin root xpublic key.

        >>> from swap.providers.bitcoin.wallet import Wallet
        >>> wallet = Wallet(network="testnet")
        >>> wallet.from_entropy("ed0802d701a033776811601dd6c5c4a9")
        >>> wallet.root_xpublic_key()
        "tpubD6NzVbkrYhZ4XpK9BpGhJuvfHJMeAggFcHCZH3NKsSbcetttiJnp184yx2cp2uJyapPQLt7LGTLUZvnKWbdgKBkvnfYjab9sH4wBmEpTZhJ"
        """

        return self._hdwallet.root_xpublic_key(encoded=encoded)

    def xprivate_key(self, encoded: bool = True) -> Optional[str]:
        """
        Get Bitcoin root xprivate key.

        :param encoded: Encoded root xprivate key, default to True.
        :type encoded: bool

        :return: str -- Bitcoin root xprivate key.

        >>> from swap.providers.bitcoin.wallet import Wallet
        >>> wallet: Wallet = Wallet(network="testnet")
        >>> wallet.from_entropy("ed0802d701a033776811601dd6c5c4a9")
        >>> wallet.from_path("m/44'/1'/0'/0/0")
        >>> wallet.xprivate_key()
        "tprv8kqWVfMdSgo9WhUAxbmL6GNW4ivePvEZBu8QiiRfMXbVDgnHx16vndnAsv7Uds4iFvjMpdJiB6q6hhh753fRb89XFjHGjYJ8BsMZGv3RTKz"
        """

        return self._hdwallet.xprivate_key(encoded=encoded)

    def xpublic_key(self, encoded: bool = True) -> Optional[str]:
        """
        Get Bitcoin xpublic key.

        :param encoded: Encoded xprivate key, default to True.
        :type encoded: bool

        :return: str -- Bitcoin xpublic key.

        >>> from swap.providers.bitcoin.wallet import Wallet
        >>> wallet: Wallet = Wallet(network="testnet")
        >>> wallet.from_entropy("ed0802d701a033776811601dd6c5c4a9")
        >>> wallet.from_path("m/44'/1'/0'/0/0")
        >>> wallet.xpublic_key()
        "tpubDHXYe5Psb4UpQAVxrFRvVg2cdkSaZFRTmCjC1ETxmoPt4B34aPvWy8Q343tUsTaCQCiSJVpzgyP1NQ3mffY7oF6u6cJ6Csx3AhgFFLVoUBu"
        """

        return self._hdwallet.xpublic_key(encoded=encoded)

    def uncompressed(self, compressed: Optional[str] = None) -> str:
        """
        Get Bitcoin Uncompressed public key.

        :param compressed: Compressed public key, default to ``None``.
        :type compressed: str

        :return: str -- Bitcoin Uncompressed public key.

        >>> from swap.providers.bitcoin.wallet import Wallet
        >>> wallet: Wallet = Wallet(network="testnet")
        >>> wallet.from_entropy("ed0802d701a033776811601dd6c5c4a9")
        >>> wallet.from_path("m/44'/1'/0'/0/0")
        >>> wallet.uncompressed()
        "f4206f9c6d35f50b3b05edc13118ab64d27959d0b7412638bfea5d132b3fb36c6d9515384318aab7fc4d15d5a1ed7999c12ae6b9d1fe11979309630d201ba632"
        """

        return self._hdwallet.uncompressed(compressed=compressed)

    def compressed(self, uncompressed: Optional[str] = None) -> str:
        """
        Get Bitcoin Compressed public key.
        
        :param uncompressed: Uncompressed public key, default to ``None``.
        :type uncompressed: str

        :return: str -- Bitcoin Compressed public key.

        >>> from swap.providers.bitcoin.wallet import Wallet
        >>> wallet: Wallet = Wallet(network="testnet")
        >>> wallet.from_entropy("ed0802d701a033776811601dd6c5c4a9")
        >>> wallet.from_path("m/44'/1'/0'/0/0")
        >>> wallet.compressed()
        "02f4206f9c6d35f50b3b05edc13118ab64d27959d0b7412638bfea5d132b3fb36c"
        """

        return self._hdwallet.compressed(uncompressed=uncompressed)

    def chain_code(self) -> str:
        """
        Get Bitcoin chain code.

        :return: str -- Bitcoin chain code.

        >>> from swap.providers.bitcoin.wallet import Wallet
        >>> wallet: Wallet = Wallet(network="testnet")
        >>> wallet.from_entropy("ed0802d701a033776811601dd6c5c4a9")
        >>> wallet.from_path("m/44'/1'/0'/0/0")
        >>> wallet.chain_code()
        "cbe00345c4cfa83dc315e52b1d5acaf2c6fce1bc8760f02696c05c3a94171304"
        """

        return self._hdwallet.chain_code()

    def private_key(self) -> str:
        """
        Get Bitcoin private key.

        :return: str -- Bitcoin private key.

        >>> from swap.providers.bitcoin.wallet import Wallet
        >>> wallet: Wallet = Wallet(network="testnet")
        >>> wallet.from_entropy("ed0802d701a033776811601dd6c5c4a9")
        >>> wallet.from_path("m/44'/1'/0'/0/0")
        >>> wallet.private_key()
        "86e4296c4b8804b952933ddf9b786a0bad1049c1d5b372e43f9336eb4ac2fcb6"
        """

        return self._hdwallet.private_key()

    def public_key(self,  compressed: bool = True, private_key: str = None) -> str:
        """
        Get Bitcoin Public key.

        :param compressed: Compressed public key, default to ``True``.
        :type compressed: bool
        :param private_key: Private key hex string, default to ``None``.
        :type private_key: str

        :return: str -- Bitcoin public key.

        >>> from swap.providers.bitcoin.wallet import Wallet
        >>> wallet: Wallet = Wallet(network="testnet")
        >>> wallet.from_entropy("ed0802d701a033776811601dd6c5c4a9")
        >>> wallet.from_path("m/44'/1'/0'/0/0")
        >>> wallet.public_key()
        "02f4206f9c6d35f50b3b05edc13118ab64d27959d0b7412638bfea5d132b3fb36c"
        """

        return self._hdwallet.public_key(compressed=compressed, private_key=private_key)

    def path(self) -> Optional[str]:
        """
        Get Bitcoin path.

        :return: str -- Bitcoin path.

        >>> from swap.providers.bitcoin.wallet import Wallet
        >>> wallet: Wallet = Wallet(network="testnet")
        >>> wallet.from_entropy("ed0802d701a033776811601dd6c5c4a9")
        >>> wallet.from_path("m/44'/1'/0'/0/0")
        >>> wallet.path()
        "m/44'/1'/0'/0/0"
        """

        return self._hdwallet.path()

    def wif(self) -> str:
        """
        Get Bitcoin important format (WIF).

        :return: str -- Bitcoin important format.

        >>> from swap.providers.bitcoin.wallet import Wallet
        >>> wallet: Wallet = Wallet(network="testnet")
        >>> wallet.from_entropy("ed0802d701a033776811601dd6c5c4a9")
        >>> wallet.from_path("m/44'/1'/0'/0/0")
        >>> wallet.wif()
        "cS6utJFQYTQEAY455hRQ5nardhCCoc2yf4M45P71ve5Dx44ag7qg"
        """

        return self._hdwallet.wif()

    def hash(self, private_key: Optional[str] = None) -> str:
        """
        Get Bitcoin public key/address hash.

        :return: str -- Bitcoin public key/address hash.

        >>> from swap.providers.bitcoin.wallet import Wallet
        >>> wallet: Wallet = Wallet(network="testnet")
        >>> wallet.from_entropy("ed0802d701a033776811601dd6c5c4a9")
        >>> wallet.from_path("m/44'/1'/0'/0/0")
        >>> wallet.hash()
        "e00ff2a640b7ce2d336860739169487a57f84b15"
        """

        return self._hdwallet.hash(private_key=private_key)

    def address(self) -> str:
        """
        Get Bitcoin address.

        :return: str -- Bitcoin address.

        >>> from swap.providers.bitcoin.wallet import Wallet
        >>> wallet: Wallet = Wallet(network="testnet")
        >>> wallet.from_entropy("ed0802d701a033776811601dd6c5c4a9")
        >>> wallet.from_path("m/44'/1'/0'/0/0")
        >>> wallet.address()
        "n1wgm6kkzMcNfAtJmes8YhpvtDzdNhDY5a"
        """

        return self._hdwallet.p2pkh_address()

    def balance(self, unit: str = config["unit"]) -> Union[int, float]:
        """
        Get Bitcoin balance.

        :param unit: Bitcoin unit, default to Satoshi.
        :type unit: str

        :return: int, float -- Bitcoin balance.

        >>> from swap.providers.bitcoin.wallet import Wallet
        >>> wallet: Wallet = Wallet(network="testnet")
        >>> wallet.from_entropy("ed0802d701a033776811601dd6c5c4a9")
        >>> wallet.from_path("m/44'/1'/0'/0/0")
        >>> wallet.balance(unit="BTC")
        0.2
        """

        if unit not in ["BTC", "mBTC", "Satoshi"]:
            raise UnitError("Invalid Bitcoin unit, choose only BTC, mBTC or Satoshi units.")
        _balance: int = get_balance(address=self.address(), network=self._network)
        return _balance if unit == "Satoshi" else \
            amount_unit_converter(amount=_balance, unit_from=f"Satoshi2{unit}")

    def utxos(self, limit: int = 15) -> list:
        """
        Get Bitcoin unspent transaction output (UTXO's).

        :param limit: Limit of UTXO's, default is 15.
        :type limit: int

        :return: list -- Bitcoin unspent transaction outputs.

        >>> from swap.providers.bitcoin.wallet import Wallet
        >>> wallet: Wallet = Wallet(network="testnet")
        >>> wallet.from_entropy("ed0802d701a033776811601dd6c5c4a9")
        >>> wallet.from_path("m/44'/1'/0'/0/0")
        >>> wallet.utxos()
        [{'index': 0, 'hash': '9d60a8b4dd16d4bf02835a21a3e9154e636ba06ad55368f36114eb7e930b35e8', 'output_index': 1, 'amount': 100000, 'script': '76a914e00ff2a640b7ce2d336860739169487a57f84b1588ac'}, {'index': 1, 'hash': '77ecb5ffe0f85454183bcab0cf1e15bfc62dc86cbdeaf374224ba03cb5cd7d29', 'output_index': 0, 'amount': 10000, 'script': '76a914e00ff2a640b7ce2d336860739169487a57f84b1588ac'}, {'index': 2, 'hash': 'e3ed50900a06990c123f3e87187009ce124cb65a46cd45eba5773fb0979fce43', 'output_index': 0, 'amount': 1797372, 'script': '76a914e00ff2a640b7ce2d336860739169487a57f84b1588ac'}]
        """

        utxos = list()
        _utxos = get_utxos(
            address=self.address(), network=self._network, limit=limit
        )
        for index, utxo in enumerate(_utxos):
            utxos.append(dict(
                index=index,
                hash=utxo["tx_hash"],
                output_index=utxo["tx_output_n"],
                amount=utxo["value"],
                script=utxo["script"]
            ))
        return utxos



#!/usr/bin/env python3

import json
import os

from swap.providers.bitcoin.htlc import HTLC
from swap.providers.bitcoin.transaction import (
    FundTransaction, WithdrawTransaction, RefundTransaction
)
from swap.providers.bitcoin.solver import (
    FundSolver, WithdrawSolver, RefundSolver
)
from swap.utils import clean_transaction_raw

# Test Values
base_path = os.path.dirname(__file__)
file_path = os.path.abspath(os.path.join(base_path, "..", "..", "values.json"))
values = open(file_path, "r")
_ = json.loads(values.read())
values.close()


def test_bitcoin_fund_transaction():

    htlc = HTLC(network=_["bitcoin"]["network"]).build_htlc(
        secret_hash=_["bitcoin"]["htlc"]["secret"]["hash"],
        recipient_address=_["bitcoin"]["wallet"]["recipient"]["address"],
        sender_address=_["bitcoin"]["wallet"]["sender"]["address"],
        endtime=_["bitcoin"]["htlc"]["endtime"]
    )

    unsigned_fund_transaction = FundTransaction(network=_["bitcoin"]["network"])

    unsigned_fund_transaction.build_transaction(
        address=_["bitcoin"]["wallet"]["sender"]["address"],
        htlc=htlc,
        amount=_["bitcoin"]["amount"],
        unit=_["bitcoin"]["unit"]
    )

    assert unsigned_fund_transaction.type() == _["bitcoin"]["fund"]["unsigned"]["type"]
    assert unsigned_fund_transaction.fee() == _["bitcoin"]["fund"]["unsigned"]["fee"]
    assert unsigned_fund_transaction.hash() == _["bitcoin"]["fund"]["unsigned"]["hash"]
    assert unsigned_fund_transaction.raw() == _["bitcoin"]["fund"]["unsigned"]["raw"]
    assert unsigned_fund_transaction.json() == _["bitcoin"]["fund"]["unsigned"]["json"]
    assert unsigned_fund_transaction.transaction_raw() == clean_transaction_raw(
        transaction_raw=_["bitcoin"]["fund"]["unsigned"]["transaction_raw"]
    )

    signed_fund_transaction = unsigned_fund_transaction.sign(
        solver=FundSolver(
            xprivate_key=_["bitcoin"]["wallet"]["sender"]["root_xprivate_key"],
            path=_["bitcoin"]["wallet"]["sender"]["derivation"]["path"],
            account=_["bitcoin"]["wallet"]["sender"]["derivation"]["account"],
            change=_["bitcoin"]["wallet"]["sender"]["derivation"]["change"],
            address=_["bitcoin"]["wallet"]["sender"]["derivation"]["address"]
        )
    )

    assert signed_fund_transaction.type() == _["bitcoin"]["fund"]["signed"]["type"]
    assert signed_fund_transaction.fee() == _["bitcoin"]["fund"]["signed"]["fee"]
    assert signed_fund_transaction.hash() == _["bitcoin"]["fund"]["signed"]["hash"]
    assert signed_fund_transaction.raw() == _["bitcoin"]["fund"]["signed"]["raw"]
    assert signed_fund_transaction.json() == _["bitcoin"]["fund"]["signed"]["json"]
    assert signed_fund_transaction.transaction_raw() == clean_transaction_raw(
        transaction_raw=_["bitcoin"]["fund"]["signed"]["transaction_raw"]
    )


def test_bitcoin_withdraw_transaction():

    unsigned_withdraw_transaction = WithdrawTransaction(network=_["bitcoin"]["network"])

    unsigned_withdraw_transaction.build_transaction(
        address=_["bitcoin"]["wallet"]["recipient"]["address"],
        transaction_hash=_["bitcoin"]["transaction_hash"]
    )

    assert unsigned_withdraw_transaction.type() == _["bitcoin"]["withdraw"]["unsigned"]["type"]
    assert unsigned_withdraw_transaction.fee() == _["bitcoin"]["withdraw"]["unsigned"]["fee"]
    assert unsigned_withdraw_transaction.hash() == _["bitcoin"]["withdraw"]["unsigned"]["hash"]
    assert unsigned_withdraw_transaction.raw() == _["bitcoin"]["withdraw"]["unsigned"]["raw"]
    assert unsigned_withdraw_transaction.json() == _["bitcoin"]["withdraw"]["unsigned"]["json"]
    assert unsigned_withdraw_transaction.transaction_raw() == clean_transaction_raw(
        transaction_raw=_["bitcoin"]["withdraw"]["unsigned"]["transaction_raw"]
    )

    signed_withdraw_transaction = unsigned_withdraw_transaction.sign(
        solver=WithdrawSolver(
            xprivate_key=_["bitcoin"]["wallet"]["recipient"]["root_xprivate_key"],
            secret_key=_["bitcoin"]["htlc"]["secret"]["key"],
            bytecode=_["bitcoin"]["htlc"]["bytecode"],
            path=_["bitcoin"]["wallet"]["recipient"]["derivation"]["path"],
            account=_["bitcoin"]["wallet"]["recipient"]["derivation"]["account"],
            change=_["bitcoin"]["wallet"]["recipient"]["derivation"]["change"],
            address=_["bitcoin"]["wallet"]["recipient"]["derivation"]["address"]
        )
    )

    assert signed_withdraw_transaction.type() == _["bitcoin"]["withdraw"]["signed"]["type"]
    assert signed_withdraw_transaction.fee() == _["bitcoin"]["withdraw"]["signed"]["fee"]
    assert signed_withdraw_transaction.hash() == _["bitcoin"]["withdraw"]["signed"]["hash"]
    assert signed_withdraw_transaction.raw() == _["bitcoin"]["withdraw"]["signed"]["raw"]
    assert signed_withdraw_transaction.json() == _["bitcoin"]["withdraw"]["signed"]["json"]
    assert signed_withdraw_transaction.transaction_raw() == clean_transaction_raw(
        transaction_raw=_["bitcoin"]["withdraw"]["signed"]["transaction_raw"]
    )


def test_bitcoin_refund_transaction():

    unsigned_refund_transaction = RefundTransaction(network=_["bitcoin"]["network"])

    unsigned_refund_transaction.build_transaction(
        address=_["bitcoin"]["wallet"]["sender"]["address"],
        transaction_hash=_["bitcoin"]["transaction_hash"]
    )

    assert unsigned_refund_transaction.type() == _["bitcoin"]["refund"]["unsigned"]["type"]
    assert unsigned_refund_transaction.fee() == _["bitcoin"]["refund"]["unsigned"]["fee"]
    assert unsigned_refund_transaction.hash() == _["bitcoin"]["refund"]["unsigned"]["hash"]
    assert unsigned_refund_transaction.raw() == _["bitcoin"]["refund"]["unsigned"]["raw"]
    assert unsigned_refund_transaction.json() == _["bitcoin"]["refund"]["unsigned"]["json"]
    assert unsigned_refund_transaction.transaction_raw() == clean_transaction_raw(
        transaction_raw=_["bitcoin"]["refund"]["unsigned"]["transaction_raw"]
    )

    signed_refund_transaction = unsigned_refund_transaction.sign(
        solver=RefundSolver(
            xprivate_key=_["bitcoin"]["wallet"]["sender"]["root_xprivate_key"],
            bytecode=_["bitcoin"]["htlc"]["bytecode"],
            endtime=_["bitcoin"]["htlc"]["endtime"],
            path=_["bitcoin"]["wallet"]["sender"]["derivation"]["path"],
            account=_["bitcoin"]["wallet"]["sender"]["derivation"]["account"],
            change=_["bitcoin"]["wallet"]["sender"]["derivation"]["change"],
            address=_["bitcoin"]["wallet"]["sender"]["derivation"]["address"]
        )
    )

    assert signed_refund_transaction.type() == _["bitcoin"]["refund"]["signed"]["type"]
    assert signed_refund_transaction.fee() == _["bitcoin"]["refund"]["signed"]["fee"]
    assert signed_refund_transaction.hash() == _["bitcoin"]["refund"]["signed"]["hash"]
    assert signed_refund_transaction.raw() == _["bitcoin"]["refund"]["signed"]["raw"]
    assert signed_refund_transaction.json() == _["bitcoin"]["refund"]["signed"]["json"]
    assert signed_refund_transaction.transaction_raw() == clean_transaction_raw(
        transaction_raw=_["bitcoin"]["refund"]["signed"]["transaction_raw"]
    )
